import sys, glob, os
import argparse
import numpy as np
import cv2
from cv2 import aruco
import matplotlib.pyplot as plt


#sqWidth = 11 # number of squares width
#sqHeight = 8 # number of squares height
#checkerSquareSize = 0.070 #0.024 # size of the checker square in (meter)
#markeSquareSize = 0.052 #0.018 # size of the marker square in (meter)

def saveCameraParams(filename,imageSize,cameraMatrix,distCoeffs,totalAvgErr):

	print(cameraMatrix)

	calibration = {'camera_matrix': cameraMatrix.tolist(),'distortion_coefficients': distCoeffs.tolist()}

	calibrationData = dict(
			 image_width = imageSize[0],
			 image_height = imageSize[1],
			 camera_matrix = dict(
				 rows = cameraMatrix.shape[0],
				 cols = cameraMatrix.shape[1],
				 dt = 'd',
				 data = cameraMatrix.tolist(),
				 ),
			 distortion_coefficients = dict(
					 rows = disCoeffs.shape[0],
					 cols = disCoeffs.shape[1],
					 dt = 'd',
					 data = disCoeffs.tolist(),
					 ),
			 avg_reprojection_error = totalAvgErr,)

	with open(filename,'w') as outfile:
			 yaml.dump(calibrationData,outfile)

def readFileList(imgFolder, ImgPattern="*.png"):
	imgFileList = glob.glob(os.path.join(imgFolder, ImgPattern))
	# self.imgFileList = os.listdir(self.imgFolder)
	# self.imgFileList.remove('.DS_Store') # remove system database log
	#imgFileList.sort(key=lambda x:int(x[len(imgFolder) + 1: -4]))
	if len(imgFileList) == 1:
		imgFileList.sort()
	else:
		imgFileList.sort()
		imgFileSubList = imgFileList[0:len(imgFileList) - 1]
		imgFileSubList.sort(key=lambda x:int(x[len(imgFolder) + 1: -4]))
		imgFileList[0:len(imgFileList) - 1] = imgFileSubList
		#print(imgFileSubList)
	#print(imgFileList)
	return imgFileList

def calibration(imgFolder, imgDistortFolder, calibFname, sqWidth, sqHeight, checkerSquareSize, markeSquareSize):

	imgFileList = readFileList(imgFolder)

	allCorners = [] #all Charuco Corners
	allIds = [] #all Charuco Ids
	decimator = 0
	#cameraMatrix = np.array([])
	#disCoeffs = np.array([])
	#viewErrors = np.array([])
	dictionary = aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250) # maker sure you use the right dictionary
	board = cv2.aruco.CharucoBoard_create(sqWidth,sqHeight,checkerSquareSize,markeSquareSize,dictionary)

	for i in imgFileList:

			# print("reading %s" % i)
			img = cv2.imread(i)
			gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
			cornerRefine = cv2.aruco.CORNER_REFINE_CONTOUR

			[markerCorners,markerIds,rejectedImgPoints] = cv2.aruco.detectMarkers(gray,dictionary)#, parameters=cv2.aruco.CORNER_REFINE_CONTOUR)


			if len(markerCorners)>0:
					[ret,charucoCorners,charucoIds] = cv2.aruco.interpolateCornersCharuco(markerCorners,markerIds,gray,board)
					if charucoCorners is not None and charucoIds is not None and len(charucoCorners)>3:# and decimator%3==0:
							allCorners.append(charucoCorners)
							allIds.append(charucoIds)

					cv2.aruco.drawDetectedMarkers(img,markerCorners,markerIds, [0, 255, 0])
					cv2.aruco.drawDetectedCornersCharuco(img,charucoCorners,charucoIds, [0, 0, 255])

					#for corner in allCorners:
					#    cv2.circle(img,(corner[0][0], corner[0][0]),50,(255,255,255))
			width = 960
			height = int(img.shape[0]*960/img.shape[1])
			smallimg = cv2.resize(img,(width,height))
			#plt.figure()
			
			#plt.imshow(smallimg)
			imgSave = os.path.join(os.path.join(imgFolder, 'calibResults'), str(decimator) + '.png') 
			cv2.imwrite(imgSave, smallimg)
			#plt.savefig('tessstttyyy.png', dpi=100)
			#plt.show()
			
			decimator+=1
	print("NumImg:", len(allCorners))
	imsize = img.shape
	print(imsize)
	#try Calibration
	try:
		#, aa, bb, viewErrors
			[ret, cameraMatrix, disCoeffs, rvecs, tvecs, _, _, perViewErrors] = cv2.aruco.calibrateCameraCharucoExtended(allCorners,allIds,board,(imsize[0], imsize[1]),
				None,None,flags=(cv2.CALIB_RATIONAL_MODEL), criteria=(cv2.TERM_CRITERIA_EPS & cv2.TERM_CRITERIA_COUNT, 10000, 1e-9))
			print("Reprojection Error:" ,ret)
			print("Camera Intrinsic Matrix:" ,cameraMatrix)
			#print("per View Errors:" ,perViewErrors)
			print("Distortion Coeffs:" ,disCoeffs)
			#print("Rvecs:", rvecs)
			#np.savez(calibFname,ret=ret,mtx=cameraMatrix,dist=disCoeffs,rvecs=rvecs,tvecs=tvecs)
			#saveCameraParams(args.file,imsize,cameraMatrix,disCoeffs,ret)
			#try to undistort the images
			#please change the folder of images to be undistorted
			
			#imgDistortFolder = '/Users/li/Documents/Deflectometry/rcalibration0418/large1117_3'
			imgDistortFilelist = readFileList(imgDistortFolder)
			img_num = 0
			for j in imgDistortFilelist:

				imgDistort = cv2.imread(j)
				imgDistortgray = cv2.cvtColor(imgDistort, cv2.COLOR_BGR2GRAY)
			#print(imgDistort.shape)
			#h, w = imgDistort.shape[:,2]
				h = imgDistort.shape[0]
				w = imgDistort.shape[1]
				newcameramtx,roi = cv2.getOptimalNewCameraMatrix(cameraMatrix, disCoeffs, (w, h), 1, (w, h))
				#print('image cropped', j)
			#undistort
				#dst = cv2.undistort(imgDistort, cameraMatrix, disCoeffs, None, newcameramtx)
				dst = cv2.undistort(imgDistort, cameraMatrix, disCoeffs, None)
				print('image undistorted', j)
				x,y,w,h = roi
				#dst = dst[y:y + h, x:x + w]
				imgDistortFilename = os.path.join(imgDistortFolder,'undistort', str(img_num) + '.png')
				cv2.imwrite(imgDistortFilename, dst)
				#print('image saved', j)
				img_num = img_num + 1
			

	except ValueError as e:
			print(e)
	except NameError as e:
			print(e)
	except AttributeError as e:
			print(e)
	except:
			print("calibrateCameraCharuco fail:" , sys.exc_info()[0])

	print("Intrinsic calibration complete!")
	return ret, cameraMatrix, disCoeffs

if __name__ == "__main__":

	parser = argparse.ArgumentParser()

	parser.add_argument("-f", "--file", help="ouput calibration filename",default="calibration.npz")
	#parser.add_argument("-s", "--size", help="size of squares in meters",type=float, default="0.035")
	parser.add_argument('-p', "--path", type=str, help='path of images for calibration')
	args = parser.parse_args()
	calibration(os.path.normpath(args.path), os.path.join(args.path, args.file))






import sys
import glob
import os
import numpy as np
import cv2
from cv2 import aruco
from skimage import color


class IntrinsicCalibration:
    def __init__(self, dict_aruco=cv2.aruco.DICT_4X4_250, sqWidth=11, sqHeight=8, checkerSquareSize=0.022,
                 markerSquareSize=0.016, criteria_eps=1e-9, criteria_count=10000):
        # Initialize ChArUco Board size values: default DINA4
        # Visit: https://calib.io/pages/camera-calibration-pattern-generator
        # To create: calib.io_charuco_279x215_8x11_24_DICT_4X4.pdf
        # Initialize dictionary, see more in OpenCV
        self.dict = dict_aruco
        # Amount of squares in width
        self.sqWidth = sqWidth
        # Amount of squares in heights
        self.sqHeight = sqHeight
        # Size of checker square on printed ChArUco board in meter
        self.checkerSquareSize = checkerSquareSize
        # Size of marker square on printed ChArUco board in meter
        self.markerSquareSize = markerSquareSize
        self.criteria_eps = criteria_eps
        self.criteria_count = criteria_count
        self.cameraMatrix = None
        self.distortionCoeff = None

    @staticmethod
    def readFileList(imgFolder, imgPattern):
        # Read all PNG files in folder
        imgFileList = glob.glob(os.path.join(imgFolder, imgPattern))
        imgFileList.sort()
        return imgFileList

    def calibration(self, imgFolder, imgPattern):
        if not os.path.exists(imgFolder):
            os.mkdir(imgFolder)
        # Retrieve Images
        imgFileList = self.readFileList(imgFolder, imgPattern=imgPattern)
        # All Charuco Corners
        allCorners = []
        # All Charuco Ids
        allIds = []
        decimator = 0
        # Retrieve dictionary
        dictionary = aruco.getPredefinedDictionary(self.dict)
        board = cv2.aruco.CharucoBoard_create(self.sqWidth, self.sqHeight, self.checkerSquareSize,
                                              self.markerSquareSize, dictionary)
        # Loop through images
        for i in imgFileList:
            print("Reading %s" % i)
            # Load image to grayscale
            if imgPattern == "*.PNG" or imgPattern == "*.png":
                img = cv2.imread(i)
                gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            elif imgPattern == "*.npy":
                img = np.load(i)
                if len(img.shape) > 2:
                    gray = color.rgb2gray(img)
                else:
                    gray = img
            else:
                print("Specify image pattern: PNG and npy are supported")
            # Detect markers
            [markerCorners, markerIds, rejectedImgPoints] = cv2.aruco.detectMarkers(gray, dictionary)
            # Draw markers
            if len(markerCorners) > 0:
                [ret, charucoCorners, charucoIds] = cv2.aruco.interpolateCornersCharuco(markerCorners, markerIds, gray,
                                                                                        board)
                if charucoCorners is not None and charucoIds is not None and len(charucoCorners) > 3:
                    allCorners.append(charucoCorners)
                    allIds.append(charucoIds)

                cv2.aruco.drawDetectedMarkers(img, markerCorners, markerIds, [0, 255, 0])
                cv2.aruco.drawDetectedCornersCharuco(img, charucoCorners, charucoIds, [0, 0, 255])

            cv2.namedWindow('Frame', cv2.WINDOW_NORMAL)
            cv2.imshow('Frame', img)
            cv2.waitKey(0)  # any key
            decimator += 1
        print("NumImg:", len(allCorners))
        # Try Calibration
        try:
            # Calibrate camera
            [ret, cameraMatrix, disCoeffs, rvecs, tvecs, _, _,
             perViewErrors] = cv2.aruco.calibrateCameraCharucoExtended(
                allCorners, allIds, board, (img.shape[0], img.shape[1]),
                None, None, flags=cv2.CALIB_RATIONAL_MODEL,
                criteria=(cv2.TERM_CRITERIA_EPS & cv2.TERM_CRITERIA_COUNT, self.criteria_count, self.criteria_eps))
            # Print computed calibration results
            print("Rep Error:", ret)
            print("Camera Matrix:", cameraMatrix)
            print("Per View Errors:", perViewErrors)
            print("Distortion Coefficients:", disCoeffs)
            print("R vecs:", rvecs)
            # Save calibration results in dedicated folder for numpy data of calibration
            np.savez('CalibrationNumpyData/intrinsic_calibration.npz', ret=ret, mtx=cameraMatrix, dist=disCoeffs,
                     rvecs=rvecs, tvecs=tvecs)
            # Set class attributes
            self.cameraMatrix = cameraMatrix
            self.distortionCoeff = disCoeffs
        except ValueError as e:
            print(e)
        except NameError as e:
            print(e)
        except AttributeError as e:
            print(e)
        except:
            print("calibrateCameraCharuco fail:", sys.exc_info()[0])
        # Close windows
        print("Press any key on window to exit")
        cv2.waitKey(0)  # any key
        cv2.destroyAllWindows()
        cv2.waitKey(1)
        cv2.waitKey(1)

    def undistort(self, distImgFolder, imgPattern):
        # Load calibration, if necessary
        if self.cameraMatrix is None:
            self.load_calibration_data()
        # Undistort the images
        imgDistortFolder = distImgFolder
        imgDistortFilelist = self.readFileList(imgDistortFolder, imgPattern=imgPattern)
        img_num = 0
        # Loop through images to undistort
        for j in imgDistortFilelist:
            # Load specified image file pattern
            if imgPattern == "*.PNG" or imgPattern == "*.png":
                imgDistort = cv2.imread(j)
            elif imgPattern == "*.npy":
                imgDistort = np.load(j)
            else:
                print("Specify image pattern: PNG and npy are supported")
            h = imgDistort.shape[0]
            w = imgDistort.shape[1]
            # TODO: Ask Yunhao why optimal camera matrix is not used
            newcameramtx, roi = cv2.getOptimalNewCameraMatrix(self.cameraMatrix, self.distortionCoeff,
                                                              (w, h), 1, (w, h))
            # Undistort
            dst = cv2.undistort(imgDistort, self.cameraMatrix, self.distortionCoeff, None)
            # Save image in specified image pattern in destination folder - by default captured image/npy
            if imgPattern == "*.PNG" or imgPattern == "*.png":
                imgSavePng = os.path.join(distImgFolder, '/capture_', str(img_num) + '.PNG')
                cv2.imwrite(imgSavePng, dst)
            elif imgPattern == "*.npy":
                imgSaveNpy = os.path.join(distImgFolder, '/capture_', str(img_num) + '.npy')
                np.save(imgSaveNpy, dst)
            else:
                print("Specify image pattern: .PNG and .npy are supported")
            img_num = img_num + 1

    def load_calibration_data(self):
        # Check if calibration file already exists
        if os.path.exists('CalibrationNumpyData/intrinsic_calibration.npz'):
            # Load camera matrix and distortion coefficients
            data = np.load('CalibrationNumpyData/intrinsic_calibration.npz')
            self.cameraMatrix = data['mtx']
            self.distortionCoeff = data['dist']
        else:
            print("Capture and calibrate camera first")



from abc import ABC
#from Reconstructions import Mesh
from GhostScan.ImageProcessing import ImageProcessing
#import wavepy
import matplotlib.pyplot as plt
import numpy as np
from skimage import color
import cv2
from tifffile import imsave
#import wavepy

np.seterr(divide='ignore', invalid='ignore')


class GradientIlluminationReconstruction(ImageProcessing, ABC):

    def __init__(self, capture_path='CapturedNumpyData/capture_%i.npy', n=2):
        super().__init__(capture_path)
        # Number of gradient images per viewing direction
        self.n = n
        self.frames_x = None
        self.frames_y = None
        self.frame_reflectivity = None
        self.diff_x = None
        self.diff_y = None
        self.magH = None
        self.magV = None
        self.normals = None
        self.depth = None
        self.albedo = None

    def loadData(self):
        #  Load first frame for dimensions and expand along 2nd axis to create a 3D array for all phase shifts
        self.frames_x = np.load(self.path % 0)
        self.frames_y = np.load(self.path % self.n)
        # RGB frames
        if len(self.frames_x.shape) == 3 or len(self.frames_y.shape) == 3:
            for i in range(1, self.n):
                self.frames_x = np.stack((self.frames_x, np.load(self.path % i)), axis=3)
                self.frames_y = np.stack((self.frames_y, np.load(self.path % (i + self.n))), axis=3)
            self.frame_reflectivity = np.load(self.path % (self.n * 2))
        # Grayscale frames
        else:
            self.frames_x = np.expand_dims(self.frames_x, axis=2)
            self.frames_y = np.expand_dims(self.frames_y, axis=2)
            for i in range(1, self.n):
                self.frames_x = np.dstack((self.frames_x, np.load(self.path % i)))
                self.frames_y = np.dstack((self.frames_y, np.load(self.path % (i + self.n))))
            self.frame_reflectivity = np.load(self.path % (self.n * 2))

    def saveTiff(self):
        # Saves normals and captured images in tiff file format
        cv2.imwrite('normals.tif',
                    cv2.cvtColor(np.array((self.normals + 1) / 2.0 * 255, dtype=np.uint8), cv2.COLOR_RGB2BGR))
        for i in range(self.n):
            imsave('captureX_%i.tif' % i, self.frames_x[:, :, i].astype(np.float16))
            imsave('captureY_%i.tif' % i, self.frames_y[:, :, i].astype(np.float16))

    def computeAlbedo(self):
        # Average over all frames to retrieve an albedo estimation
        # RGB frames
        if len(self.frames_x.shape) == 4 or len(self.frames_y.shape) == 4:
            stack = np.concatenate((self.frames_x, self.frames_y), axis=3)
            self.albedo = np.mean(stack, axis=3)
        # Grayscale
        else:
            stack = np.dstack((self.frames_x, self.frames_y))
            self.albedo = np.mean(stack, axis=2)

        alb = (self.albedo - np.min(self.albedo)) / (np.max(self.albedo)-np.min(self.albedo))
        alb *= 255.0
        # Save the albedo as PNG
        # RBG
        if len(self.frames_x.shape) == 4 or len(self.frames_y.shape) == 4:
            alb = cv2.cvtColor(np.array(alb, dtype=np.uint8), cv2.COLOR_BGR2RGB)
            cv2.imwrite('Results/albedo.PNG', alb)
        # Grayscale
        else:
            cv2.imwrite('Results/albedo.PNG', np.array(alb, dtype=np.uint8))

    def computeNormalMapLookUp(self):
        # Ineffective
        if self.n > 2:
            print("Gradient shifting requires, two gradients per viewing direction. Average your data to fit.")
        if self.frames_y is None or self.frames_x is None:
            print("Load data first, using loadData")
        """
        # Get overall maximum value
        max_norm = max(np.max(self.frames_x), np.max(self.frames_y))
        # It is required to remove the mean from the captured images
        self.frames_x = (self.frames_x / max_norm)
        self.frames_y = (self.frames_y / max_norm)
        """
        # Instead of normalizing frame by frame we normalize by the last captured image
        # this image is taken under constant maximum illumination and can therefore correct
        # for different colors and normalize pixel by pixel
        # RGB to grayscale
        if len(self.frames_x.shape) == 4 or len(self.frames_y.shape) == 4:
            frames_x_n = np.dstack(
                (color.rgb2gray(self.frames_x[..., 0]), color.rgb2gray(self.frames_x[..., 1]))).astype(np.float64)
            frames_y_n = np.dstack(
                (color.rgb2gray(self.frames_y[..., 0]), color.rgb2gray(self.frames_y[..., 1]))).astype(np.float64)
            frame_r_n = color.rgb2gray(self.frame_reflectivity).astype(np.float64) + 1e-9
        # Grayscale
        else:
            frames_x_n = self.frames_x.astype(np.float64)
            frames_y_n = self.frames_y.astype(np.float64)
            frame_r_n = self.frame_reflectivity.astype(np.float64) + 1e-9

        for i in range(self.n):
            frames_x_n[..., i] = frames_x_n[..., i] / frame_r_n
            frames_y_n[..., i] = frames_y_n[..., i] / frame_r_n

        """
        lookup = np.load('CalibrationNumpyData/lookup.npy')
        l_x_0 = np.abs(lookup - np.repeat(np.expand_dims(frames_x_n[..., 0], axis=2), lookup.shape[2], axis=2))
        l_x_1 = np.abs(lookup - np.repeat(np.expand_dims(frames_x_n[..., 1], axis=2), lookup.shape[2], axis=2))
        l_y_0 = np.abs(lookup - np.repeat(np.expand_dims(frames_y_n[..., 0], axis=2), lookup.shape[2], axis=2))
        l_y_1 = np.abs(lookup - np.repeat(np.expand_dims(frames_y_n[..., 1], axis=2), lookup.shape[2], axis=2))
        l_x_0 = np.argmin(l_x_0, axis=2)
        l_x_1 = np.argmin(l_x_1, axis=2)
        l_y_0 = np.argmin(l_y_0, axis=2)
        l_y_1 = np.argmin(l_y_1, axis=2)
        frames_x_n[..., 0] = l_x_0
        frames_x_n[..., 1] = l_x_1
        frames_y_n[..., 0] = l_y_0
        frames_y_n[..., 1] = l_y_1
        np.save('CapturedNumpyData/capture_lookupx0.npy', l_x_0)
        np.save('CapturedNumpyData/capture_lookupx1.npy', l_x_1)
        np.save('CapturedNumpyData/capture_lookupy0.npy', l_y_0)
        np.save('CapturedNumpyData/capture_lookupy1.npy', l_y_1)
        """
        # Compute difference of gradient illuminations in opposing directions
        self.diff_x = frames_x_n[..., 1] - frames_x_n[..., 0]
        self.diff_y = frames_y_n[..., 1] - frames_y_n[..., 0]
        # Compute normals
        z = np.sqrt(1 - np.square(self.diff_x) - np.square(self.diff_y))
        norm = np.sqrt(np.square(self.diff_x) + np.square(self.diff_y) + np.square(z))
        x = self.diff_x / norm
        y = self.diff_y / norm
        z /= norm
        # Create normals array
        self.normals = np.stack((x, y, z), axis=2)
        # Save as PNG file
        cv2.imwrite('Results/normals.PNG', cv2.cvtColor(np.array((self.normals + 1) / 2.0 * 255, dtype=np.uint8),
                                                        cv2.COLOR_RGB2BGR))

    def computeNormalMapSingle(self, gamma):
        if self.n > 2:
            print("Gradient shifting requires, two gradients per viewing direction. Average your data to fit.")
        if self.frames_y is None or self.frames_x is None:
            print("Load data first, using loadData")
        # Instead of normalizing frame by frame we normalize by the last captured image
        # this image is taken under constant maximum illumination and can therefore correct
        # for different colors and normalize pixel by pixel
        # RGB to grayscale

        if len(self.frames_x.shape) == 4 or len(self.frames_y.shape) == 4:
            frames_x_n = np.dstack(
                (color.rgb2gray(self.frames_x[..., 0]), color.rgb2gray(self.frames_x[..., 1]))).astype(np.float64)
            frames_y_n = np.dstack(
                (color.rgb2gray(self.frames_y[..., 0]), color.rgb2gray(self.frames_y[..., 1]))).astype(np.float64)
            frame_r_n = color.rgb2gray(self.frame_reflectivity).astype(np.float64) + 1e-9
        # Grayscale
        else:
            frames_x_n = self.frames_x.astype(np.float64)**gamma
            frames_y_n = self.frames_y.astype(np.float64)**gamma
            frame_r_n = self.frame_reflectivity.astype(np.float64)**gamma

        # Get overall maximum value
        max_norm = max(np.max(frames_x_n), np.max(frames_y_n))
        # It is required to remove the mean from the captured images
        frames_x_n = (frames_x_n / max_norm)
        frames_y_n = (frames_y_n / max_norm)

        # Compute difference of gradient illuminations in opposing directions
        self.diff_x = frames_x_n[..., 1] - frames_x_n[..., 0]
        self.diff_y = frames_y_n[..., 1] - frames_y_n[..., 0]
        # Compute normals
        z = np.sqrt(1 - np.square(self.diff_x) - np.square(self.diff_y))
        norm = np.sqrt(np.square(self.diff_x) + np.square(self.diff_y) + np.square(z))
        x = self.diff_x / norm
        y = self.diff_y / norm
        z /= norm
        # Create normals array
        self.normals = np.stack((x, y, z), axis=2)
        # Save as PNG file
        cv2.imwrite('Results/normals.PNG', cv2.cvtColor(np.array((self.normals + 1) / 2.0 * 255, dtype=np.uint8),
                                                        cv2.COLOR_RGB2BGR))
        np.save('Results/normals.npy', self.normals)

    def computeNormalMapRadiance(self, gamma):
        if self.n > 2:
            print("Gradient shifting requires, two gradients per viewing direction. Average your data to fit.")
        if self.frames_y is None or self.frames_x is None:
            print("Load data first, using loadData")
        # Instead of normalizing frame by frame we normalize by the last captured image
        # this image is taken under constant maximum illumination and can therefore correct
        # for different colors and normalize pixel by pixel
        # RGB to grayscale
        if len(self.frames_x.shape) == 4 or len(self.frames_y.shape) == 4:
            frames_x_n = np.dstack(
                (color.rgb2gray(self.frames_x[..., 0]), color.rgb2gray(self.frames_x[..., 1]))).astype(np.float64)
            frames_y_n = np.dstack(
                (color.rgb2gray(self.frames_y[..., 0]), color.rgb2gray(self.frames_y[..., 1]))).astype(np.float64)
            frame_r_n = color.rgb2gray(self.frame_reflectivity).astype(np.float64) + 1e-9
        # Grayscale
        else:
            frames_x_n = self.frames_x.astype(np.float64)**gamma
            frames_y_n = self.frames_y.astype(np.float64)**gamma
            frame_r_n = self.frame_reflectivity.astype(np.float64)**gamma + 1e-9

        for i in range(self.n):
            frames_x_n[..., i] = frames_x_n[..., i] #/ frame_r_n
            frames_y_n[..., i] = frames_y_n[..., i] #/ frame_r_n

        # Get overall maximum value
        max_norm = max(np.max(frames_x_n), np.max(frames_y_n))
        # It is required to remove the mean from the captured images
        frames_x_n = (frames_x_n / max_norm)
        frames_y_n = (frames_y_n / max_norm)

        # Compute difference of gradient illuminations in opposing directions
        self.diff_x = frames_x_n[..., 1] - frames_x_n[..., 0]
        self.diff_y = frames_y_n[..., 1] - frames_y_n[..., 0]
        # Compute normals
        z = np.sqrt(1 - np.square(self.diff_x) - np.square(self.diff_y))
        norm = np.sqrt(np.square(self.diff_x) + np.square(self.diff_y) + np.square(z))
        x = self.diff_x / norm
        y = self.diff_y / norm
        z /= norm
        # Create normals array
        self.normals = np.stack((x, y, z), axis=2)
        # Save as PNG file
        cv2.imwrite('Results/normals.PNG', cv2.cvtColor(np.array((self.normals + 1) / 2.0 * 255, dtype=np.uint8),
                                                        cv2.COLOR_RGB2BGR))
        np.save('Results/normals.npy', self.normals)

    def computePointCloud(self, crop=((0, 0), (0, 0))):
        # Crop by ((start_of_crop_x, height_of_crop_x), (start_of_crop_y, height_of_crop_y))
        # IMPORTANT: Note the second tuple parameters are the cropping height/width, not the x/y crop end positions !!
        if crop[0][1] == 0 or crop[0][1] == 0:
            height = self.frames_x.shape[0]
            width = self.frames_x.shape[1]
        else:
            height = crop[0][1]
            width = crop[1][1]
        crop_x0 = crop[0][0]
        crop_y0 = crop[1][0]
        crop_x1 = crop_x0 + height
        crop_y1 = crop_y0 + width
        # Create mesh object and set normals, albedo, and compute depth
        meshGI = Mesh.Mesh("GradientIllumination", width, height)
        meshGI.setNormal(self.normals[crop_y0:crop_y1, crop_x0:crop_x1, ...])
        meshGI.setTexture(self.albedo[crop_y0:crop_y1, crop_x0:crop_x1, ...])
        self.depth = meshGI.setDepth()
        depth_png = (self.depth - np.min(self.depth)) / (np.max(self.depth) - np.min(self.depth))
        depth_png *= 255.0
        cv2.imwrite('Results/depth.PNG', depth_png)
        # Create mesh obj to export
        meshGI.exportOBJ("Results/", True)

    @staticmethod
    def normalizeRGBNormalsPQ(normals_red, normals_green, normals_blue):
        # Normals under red light
        p_red = -normals_red[..., 0] / normals_red[..., 2]
        q_red = -normals_red[..., 1] / normals_red[..., 2]
        p_red = np.nan_to_num(p_red)
        q_red = np.nan_to_num(q_red)
        # Normals under green light
        p_green = -normals_green[..., 0] / normals_green[..., 2]
        q_green = -normals_green[..., 1] / normals_green[..., 2]
        p_green = np.nan_to_num(p_green)
        q_green = np.nan_to_num(q_green)
        # Normals under blue light
        p_blue = -normals_blue[..., 0] / normals_blue[..., 2]
        q_blue = -normals_blue[..., 1] / normals_blue[..., 2]
        p_blue = np.nan_to_num(p_blue)
        q_blue = np.nan_to_num(q_blue)
        return p_red, q_red, p_green, q_green, p_blue, q_blue

    @staticmethod
    def normalizeRGNormalsPQ(normals_red, normals_green):
        # Normals under red light
        p_red = -normals_red[..., 0] / normals_red[..., 2]
        q_red = -normals_red[..., 1] / normals_red[..., 2]
        p_red = np.nan_to_num(p_red)
        q_red = np.nan_to_num(q_red)
        # Normals under green light
        p_green = -normals_green[..., 0] / normals_green[..., 2]
        q_green = -normals_green[..., 1] / normals_green[..., 2]
        p_green = np.nan_to_num(p_green)
        q_green = np.nan_to_num(q_green)
        return p_red, q_red, p_green, q_green

    @staticmethod
    def computeSegmentationRGBDepth(normals_red, normals_green, normals_blue):
        p_red, q_red, p_green, q_green, p_blue, q_blue = GradientShiftingReconstruction.normalizeRGBNormalsPQ(
            normals_red, normals_green, normals_blue)
        t1 = (p_red+(p_green+p_blue)/2)/2
        t2 = (q_red + (q_green + q_blue) / 2) / 2
        z = wavepy.surface_from_grad.frankotchellappa(t1, t2, False)
        z_all = z.real
        zr = wavepy.surface_from_grad.frankotchellappa(p_red, q_red, False)
        z_red = zr.real
        z_real = np.minimum(z_all, z_red)
        return z_real, t1, t2

    @staticmethod
    def computeGradientSpreadRGBDepth(normals_red, normals_green):
        # Normalize by z component to obtain P and Q
        p_red, q_red, p_green, q_green = GradientShiftingReconstruction.normalizeRGNormalsPQ(
            normals_red, normals_green)
        print(p_red.shape, p_green.shape, q_red.shape, q_green.shape)
        # Fourier Transform
        p_red_fft = np.fft.fft(p_red)
        p_green_fft = np.fft.fft(p_red)
        q_red_fft = np.fft.fft(q_red)
        q_green_fft = np.fft.fft(q_green)
        """
        print(p_red_fft.shape, p_green_fft.shape, q_red_fft.shape, q_green_fft.shape)
        # Meshgrid for spatial coordinates
        x = np.linspace(1, p_red.shape[1]+1, p_red.shape[1])
        y = np.linspace(1, p_red.shape[0]+1, p_red.shape[0])
        u, v = np.meshgrid(x, y)
        print(u.shape, v.shape)
        # Enforce criterium after Dong et al.
        p_r_cost = v*p_red_fft - u*q_red_fft
        p_g_cost = v*p_green_fft - u*q_green_fft
        p_cost = np.dstack((p_r_cost, p_g_cost))
        # Find gradient that minimizes constraint
        ids = np.argmin(p_cost, axis=2)
        n_ids = np.where((ids == 0) | (ids == 1), ids ^ 1, ids)
        opt_fft_p = n_ids * p_red + ids * p_green
        opt_fft_q = n_ids * q_red + ids * q_green
        """
        """
        p_stack = np.dstack((p_red_fft, q_red_fft))
        q_stack = np.dstack((q_red_fft, q_red_fft))
        print(p_stack.shape, q_stack.shape)
        # Take fourier transform of gradient that minimizes above mentioned
        opt_fft_p = p_stack[..., ids]
        opt_fft_q = q_stack[..., ids]
        
        print(opt_fft_p.shape, opt_fft_q.shape)
        """
        # Inverse Fourier Transform to acquire spatial gradients
        #opt_p = np.fft.ifft(opt_fft_p)
        #opt_q = np.fft.ifft(opt_fft_q)
        opt_p = np.fft.ifft(p_red_fft)
        opt_q = np.fft.ifft(q_red_fft)
        # Fanko Chelappa to integrate to depth values
        #opt_z = sfg.frankotchellappa(opt_p, opt_q, False)
        return opt_p.real, opt_q.real#, opt_z

    def highPassFilter(self):
        # Smooth filter operation
        kernel = np.ones((5, 5), np.float32) / 25
        self.normals = cv2.filter2D(self.normals, -1, kernel)



import sys, glob, os
import math
import argparse
import numpy as np
import cv2
from cv2 import aruco
from GhostScan.Calibrations.Camera2ScreenCalib import calib
# from Camera2ScreenCalib import calib
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def geoCalib(imgFile, imgUndistortFile, half_length, half_height, cameraMatrix, disCoeffs, ret, displayScaleFactor):

    # calculate cam2screen
    rc2s, tc2s = calib(imgFile, cameraMatrix, disCoeffs, half_length, half_height, displayScaleFactor)
    tc2s = tc2s[0]
    #print("new rc2s is ", rc2s)
    #print("new tc2s is ", tc2s)
    rcamera_dis = rc2s
    tcamera_dis = tc2s
    # additional section: Make sure that the shape of rC2S and tC2S is correct
    rcamera_dis = rcamera_dis[0]
    tcamera_dis = np.reshape(tcamera_dis, (3,))
    #print("rc2s is ",rcamera_dis)
    #print("tc2s is ",tcamera_dis)
    #print("intrinsic matrix is ",cameraMatrix)



    #print(intrinsic_mat.f.dist)
    allCorners = []
    allIds = []
    decimator = 0
    dictionary = aruco.getPredefinedDictionary(cv2.aruco.DICT_6X6_250)
    #board = 
    img = cv2.imread(imgFile)
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    cornerRefine = cv2.aruco.CORNER_REFINE_CONTOUR
    [markerCorners,markerIDs, rejectedImgPoints] = cv2.aruco.detectMarkers(gray, dictionary, cameraMatrix=cameraMatrix, distCoeff=disCoeffs)



    board_corner = np.array([np.array([[-half_length, -half_height, 0], [-half_length + 0.04, -half_height, 0], [-half_length + 0.04, -half_height + 0.04, 0], [-half_length, -half_height + 0.04, 0]]),
                    np.array([[half_length - 0.04, -half_height, 0], [half_length, -half_height, 0], [half_length, -half_height + 0.04, 0], [half_length - 0.04, -half_height + 0.04, 0]]),
                    np.array([[half_length - 0.04, half_height - 0.04, 0], [half_length, half_height - 0.04, 0], [half_length, half_height, 0], [half_length - 0.04, half_height, 0]]),
                    np.array([[-half_length, half_height - 0.04, 0], [-half_length + 0.04, half_height - 0.04, 0], [-half_length + 0.04, half_height, 0], [-half_length, half_height, 0]]),
                    np.array([[-0.02, -half_height, 0], [0.02, -half_height, 0], [0.02, -half_height + 0.04, 0], [-0.02, -half_height + 0.04, 0]]),
                    np.array([[-half_length, -0.02, 0], [-half_length + 0.04, -0.02, 0], [-half_length + 0.04, 0.02, 0], [-half_length, 0.02, 0]]),
                    np.array([[half_length - 0.04, -0.02, 0], [half_length, -0.02, 0], [half_length, 0.02, 0], [half_length - 0.04, 0.02, 0]]),
                    np.array([[-0.02, half_height - 0.04, 0], [0.02, half_height - 0.04, 0], [0.02, half_height, 0], [-0.02, half_height, 0]])], dtype=np.float32)

    board_id = np.array([[0], [6], [4], [2], [7], [1], [5], [3]], dtype=np.int32)
    board = cv2.aruco.Board_create(board_corner, dictionary, board_id)





    if len(markerCorners) > 0:
        allCorners.append(markerCorners)
        allIds.append(markerIDs)
        cv2.aruco.drawDetectedMarkers(img, markerCorners, markerIDs, [0,200,0])



    #cv2.aruco.drawPlanarBoard(board, img.shape)
    rvecs,tvecs,_objpoints = cv2.aruco.estimatePoseSingleMarkers(markerCorners, 0.04, cameraMatrix, disCoeffs, )
    r_mean = np.mean(rvecs,axis=0)
    t_mean = np.mean(tvecs,axis=0)

    #r_mirror = np.zeros((1,3))
    #t_mirror = np.zeros((1,3))
    r_mirror = None
    t_mirror = None
    ret, r_mirror, t_mirror, = cv2.aruco.estimatePoseBoard(markerCorners, markerIDs, board, cameraMatrix, disCoeffs, r_mirror, t_mirror)
    #print("r vec of mirror is ", r_mirror)
    #print("t vec of mirror is ", t_mirror)
    r_mirror_mat = np.zeros((3,3))
    r_mirror_mat = cv2.Rodrigues(r_mirror,dst= r_mirror_mat,jacobian=None)
    r_mirror_mat = r_mirror_mat[0]
    #print("r mirror mat is ", r_mirror_mat)
    img_axis = aruco.drawAxis(img, cameraMatrix, disCoeffs, r_mirror, t_mirror, 0.04)
    width = 960
    height = int(img.shape[0]*960/img.shape[1])
    smallimg = cv2.resize(img,(width,height))
    smallimg = cv2.cvtColor(smallimg, cv2.COLOR_BGR2RGB)
    fig = plt.figure(figsize=(15,10))
    plt.imshow(smallimg)
    # print(str(decimator) + '!!!!!!!!!!!!')
    imgSavePath = os.path.join(os.path.join(os.path.join(os.getcwd(), 'CalibrationImages'), 'Geometric'), 'geoCalibResults')
    fig.savefig(imgSavePath + '/findAllAruco' + '.png')
    plt.show()
    decimator+=1
    imsize = img.shape
    #print("size of img is ", imsize)


    #rtcam: camera coordinate system to world coordinate system
    #rtMat: world coordinate system to camera coordinate system
    rtMat = np.hstack((r_mirror_mat, t_mirror))
    rtMat = np.vstack((rtMat, np.array([[0, 0, 0, 1]])))
    rt_cam = np.linalg.inv(rtMat)
    r_cam = rt_cam[0:3, 0:3]
    t_cam = rt_cam[0:3, 3]
    #print("rtMat is", rtMat)
    #print("rtcam in world coor is ", rt_cam)

    #rtdis: display coordinate system to world coordinate system
    #rtdis_inv: world coordinate system to display coordinate system
    t_camdis = np.reshape(tcamera_dis, (tcamera_dis.shape[0],1))
    rtMat_camdis = np.hstack((rcamera_dis, t_camdis))
    rtMat_camdis = np.vstack((rtMat_camdis, np.array([0,0,0,1])))
    #rtMat_discam = np.linalg.inv(rtMat_camdis)
    #rt_dis_inv = np.matmul(rtMat, rtMat_camdis)
    rt_dis_inv = np.matmul(rtMat_camdis, rtMat)
    rt_dis = np.linalg.inv(rt_dis_inv)
    r_dis = rt_dis[0:3, 0:3]
    t_dis = rt_dis[0:3,3]
    r_dis_inv = rt_dis_inv[0:3,0:3]
    t_dis_inv = rt_dis_inv[0:3,3]

    #rtMat_camdis: camera coordinate to display coordinate system
    #rtMat_discam: display coordinate to camera coordinate system
    rtMat_discam = np.linalg.inv(rtMat_camdis)


    #print("rtscreen in world coor is ", rt_dis)
    #print("mirror in screen coor is ",rt_dis_inv)
    #print("display to camera coor is ", rtMat_discam)
    #reshape t vector
    t_cam = np.reshape(t_cam, (t_cam.shape[0], 1))
    t_dis = np.reshape(t_dis, (t_dis.shape[0], 1))
    t_dis_inv = np.reshape(t_dis_inv, (t_dis_inv.shape[0], 1))



    # read undistorted image
    img_undistort = cv2.imread(imgUndistortFile)
    gray_undistort = cv2.cvtColor(img_undistort, cv2.COLOR_BGR2GRAY)
    img_undistort_size = gray_undistort.shape
    #print(img_undistort_size)


    #cam roor mat:store arrival vectors (vector from camera to the object) in camera coordinate system
    #cam world coor mat: store arrival vectors in world coordinate system
    #camera mirror intersect mat: store intersect points between arrival vectors and mirror in world coordinate system
    # reflec mat: store reflectance vectors (vector from object to screen) in world coordinate system 
    # mirror intersect mat trans:store intersect points on mirrors at display coordinate system
    # reflect mat trans: store reflectance vectors in display coordinate system
    # display intersect mat: store display intersect points (intersect points between reflectance vectors and display) on display coordinate system
    # display intersect mat trans: store display intersect points on world coordinate system
    img_coor_mat = np.zeros((3, img_undistort_size[0], img_undistort_size[1]))
    img_coor_mat_rs = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
    cam_coor_mat = np.zeros((img_undistort_size[0], img_undistort_size[1], 3))
    cam_coor_mat_rs = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
    cam_world_coor_mat = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
    cam_mirror_intersect_mat = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
    reflect_mat = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
    mirror_intersect_mat_trans = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
    reflect_mat_trans = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
    display_intersect_mat = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
    display_intersect_mat_trans = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))

    #print(cam_coor_mat.shape)
    #print(np.linalg.inv(cameraMatrix))
    for i in range(img_undistort_size[0]):
        for j in range(img_undistort_size[1]):
            image_coor = np.array([j + 1, i + 1, 1])
            img_coor_mat[:, i, j] = image_coor

    img_coor_mat_rs = np.reshape(img_coor_mat, (3, img_undistort_size[0] * img_undistort_size[1]))
    #print(img_coor_mat_rs[:,1201])
    # calculate arrival vector
    cam_coor_mat_rs = np.matmul(np.linalg.inv(cameraMatrix), img_coor_mat_rs)
    cam_world_coor_mat = np.matmul(r_cam, cam_coor_mat_rs)
    # calculate intersect point
    scale_factor1 = -t_cam[2,0] / cam_world_coor_mat[2,:]
    scale_factor1 = np.reshape(scale_factor1, (1, scale_factor1.shape[0]))
    cam_mirror_intersect_mat = scale_factor1 * cam_world_coor_mat + np.tile(t_cam, (cam_world_coor_mat.shape[1]))
    # calculate reflectance vector in world coor system
    reflect_mat = cam_world_coor_mat - 2 * np.dot(np.array([0,0,1]), cam_world_coor_mat) * np.tile(np.array([[0],[0],[1]]), (1, cam_world_coor_mat.shape[1]))
    # calculate intersect point in display coor system
    mirror_intersect_mat_trans = np.dot(r_dis_inv, cam_mirror_intersect_mat) + np.tile(t_dis_inv, (cam_world_coor_mat.shape[1]))
    # calculate reflectance vector in display coor system
    reflect_mat_trans = np.dot(r_dis_inv, reflect_mat)
    # calculate display intersect point under display coordinate system
    scale_factor2 = -mirror_intersect_mat_trans[2,:] / reflect_mat_trans[2,:]
    display_intersect_mat = scale_factor2 * reflect_mat_trans + mirror_intersect_mat_trans
    # calculate display intersect point under world coordinate system
    display_intersect_mat_trans = np.dot(r_dis, display_intersect_mat) + np.tile(t_dis, (cam_world_coor_mat.shape[1]))

    # restore the matrix back to the same shape of undistort image
    display_intersect_mat = np.reshape(display_intersect_mat, (3, img_undistort_size[0], img_undistort_size[1]))
    display_intersect_mat_trans = np.reshape(display_intersect_mat_trans, (3, img_undistort_size[0], img_undistort_size[1]))
    cam_mirror_intersect_mat = np.reshape(cam_mirror_intersect_mat, (3, img_undistort_size[0], img_undistort_size[1]))

    #np.savetxt("displayintersect0.txt",display_intersect_mat[0,:,:])
    #np.savetxt("displayintersect1.txt",display_intersect_mat[1,:,:])

    ## this part is to display camera, mirror &projector within a unified world coordinate systems

    #reshape t vector
    t_cam = np.reshape(t_cam, (t_cam.shape[0],))
    t_dis = np.reshape(t_dis, (t_dis.shape[0],))
    t_dis_inv = np.reshape(t_dis_inv, (t_dis_inv.shape[0], ))

    #mirror corner index
    markerCornerIDs = np.array([np.argwhere(markerIDs == 0), np.argwhere(markerIDs == 2), np.argwhere(markerIDs == 4), np.argwhere(markerIDs == 6)])
    markerCornerIDs = markerCornerIDs[:,0,0]

    mirror_corner_value = np.array([markerCorners[markerCornerIDs[0]][0][0],
    markerCorners[markerCornerIDs[1]][0][3],
    markerCorners[markerCornerIDs[2]][0][2],
    markerCorners[markerCornerIDs[3]][0][1],
    markerCorners[markerCornerIDs[0]][0][0]], dtype=np.int32)
    #print("mirror corner index is", mirror_corner_value)
    mirror_corner_world = np.zeros((mirror_corner_value.shape[0],3))
    for i in range(mirror_corner_world.shape[0]):
        mirror_corner_world[i,:] = cam_mirror_intersect_mat[:, mirror_corner_value[i,1], mirror_corner_value[i,0]]
    #print("mirror corner point is", mirror_corner_world)

    #display corners in world coordinate system
    half_length_disp = 960 * displayScaleFactor
    half_height_disp = 540 * displayScaleFactor
    disp_uleft = np.array([-half_length_disp, -half_height_disp, 0])
    disp_lleft = np.array([-half_length_disp, half_height_disp, 0])
    disp_lright = np.array([half_length_disp, half_height_disp, 0])
    disp_uright = np.array([half_length_disp, -half_height_disp, 0])
    disp_uleft_world = np.reshape(np.matmul(r_dis, disp_uleft) + t_dis, (1, t_dis.shape[0]))
    disp_lleft_world = np.reshape(np.matmul(r_dis, disp_lleft) + t_dis, (1, t_dis.shape[0]))
    disp_lright_world = np.reshape(np.matmul(r_dis, disp_lright) + t_dis, (1,t_dis.shape[0]))
    disp_uright_world = np.reshape(np.matmul(r_dis, disp_uright) + t_dis, (1, t_dis.shape[0]))

    disp_corners_world = np.concatenate((disp_uleft_world, disp_lleft_world, disp_lright_world, disp_uright_world, disp_uleft_world), axis= 0)
    #print("disp corners in world coordinate is ", disp_corners_world)

    # calculate the (0,0) and z axis of camera and display (for verification)
    cam_z1 = np.matmul(r_cam, np.array([0,0,1])) + t_cam
    cam_zero = np.matmul(r_cam, np.array([0,0,0])) + t_cam
    display_z1 = np.matmul(r_dis,np.array([0,0,1])) + t_dis
    display_zero = np.matmul(r_dis,np.array([0,0,0])) + t_dis
    #cam_z = np.concatenate((np.reshape(cam_zero, (1, cam_zero.shape[0])), np.reshape(cam_z1, (1, cam_z1.shape[0]))), axis= 0)
    #display_z = np.concatenate((np.reshape(display_zero, (1,display_zero.shape[0])), np.reshape(display_z1, (1, display_z1.shape[0]))), axis= 0)
    cam_z = cam_z1 - cam_zero
    display_z = display_z1 - display_zero
    print('cam z axis in world coor is ',cam_z)
    print("cam zero point in world coor is ", cam_zero)
    print("display z axis in world coor is ", display_z)
    print("display zero in world coor is ", display_zero)
    # plot 
    fig = plt.figure(figsize=(15,10))
    ax = Axes3D(fig)
    ax.plot(mirror_corner_world[:,0], mirror_corner_world[:,1], mirror_corner_world[:,2], label= 'mirror')
    ax.plot(disp_corners_world[:,0], disp_corners_world[:,1], disp_corners_world[:,2], label= 'display')
    ax.quiver(cam_zero[0], cam_zero[1], cam_zero[2], cam_z[0], cam_z[1], cam_z[2], length= 0.4)
    ax.quiver(display_zero[0], display_zero[1], display_zero[2], display_z[0], display_z[1], display_z[2], length= 0.4)
    plt.title("Screen, Camera and Object in World Coordinate System")
    plt.legend()
    plt.show()
    imgPath = os.path.join(os.path.join(os.path.join(os.getcwd(), 'CalibrationImages'), 'Geometric'), 'geoCalibResults')
    fig.savefig(imgPath + '/Screen, Camera and Object in World Coordinate System' + '.png')

    return display_intersect_mat, cam_mirror_intersect_mat, rcamera_dis, tcamera_dis



from abc import ABC
import tkinter
from PIL import Image, ImageTk
import numpy as np
import cv2
import time
# if using an environment different to PyCharm
# import sys
# sys.path.insert(0,'..')
import warnings
warnings.filterwarnings("ignore", message="module not found")
from GhostScan.Projection import Projection


class Screen(Projection, ABC):
    def __init__(self, frequency=0):
        # Initialize internal screen fullscreen
        self.root = tkinter.Tk()
        #self.root.attributes('-fullscreen', True)
        #self.root.overrideredirect(1)
        # Get screen resolution
        w, h = self.root.winfo_screenwidth(), self.root.winfo_screenheight()
        print("Screen resolution: ", (w, h))
        resolution = (w, h)
        self.root.overrideredirect(True)
        self.root.geometry("{0}x{1}+0+0".format(w, h))
        self.root.focus_set()
        self.root.bind("<Escape>", lambda e: e.widget.quit())
        # Include quit button
        # b = tkinter.Button(self.root, text="quit", command=self.quit_and_close)
        # b.pack()
        # Create black canvas
        self.canvas = tkinter.Canvas(self.root, width=w, height=h)
        self.canvas.pack()
        self.canvas.configure(background='white')
        # Init base class
        super().__init__(resolution, frequency)
        self.count = 0
        # Create empty pattern and camera
        self.pattern = None
        self.camera = None

    def displayCalibrationPattern(self, camera, path_calib='CalibrationImages/8_24_checker.png',
                                  save_img='Geometric/geo'):
        self.camera = camera
        img = cv2.imread(path_calib)
        cv2.namedWindow('Checkerboard', cv2.WINDOW_NORMAL)
        cv2.setWindowProperty('Checkerboard', cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)
        cv2.imshow('Checkerboard', img)
        time.sleep(2)
        self.camera.getImage(name=save_img, calibration=True)
        cv2.waitKey(0)  # any key
        cv2.destroyWindow('Checkerboard')

    def displayPatterns(self, camera):
        # Displays a series of pattern, which is updated in updateCanvas
        # If you only desire to view the projection, pass on None for the input camera
        self.camera = camera
        self.root.after(100, self.updateCanvas())
        self.root.mainloop()
        self.root = None

    def setPattern(self, pattern):
        # Sets pattern to project
        self.pattern = pattern

    def updateCanvas(self):
        # Updates the pattern to project
        if self.count >= self.pattern.patterns.shape[-1]:
            # If done, quit projection
            self.root.destroy()
            return
        # Create Pil image from Np array
        modulation = self.pattern.patterns[..., self.count]*255
        pilImage = Image.fromarray(modulation.astype(np.uint8))
        image = ImageTk.PhotoImage(pilImage)
        # Set image
        self.canvas.create_image(0, 0, anchor='nw', image=image)
        self.root.update()
        time.sleep(2)
        if self.camera is None:
            print("No camera initialized.")
        else:
            # Take snapshot
            if self.camera.hdr_exposures is None:
                self.camera.getImage(name='capture_'+str(self.count))
            else:
                self.camera.getHDRImage(name='capture_'+str(self.count))
        self.count += 1
        self.root.after(100, self.updateCanvas())

    def getResolution(self):
        # Returns tuple of resolution (width x height)
        return self.resolution

    def setResolution(self, resolution):
        # Sets tuple of resolution (width x height)
        self.resolution = resolution

    def quit_and_close(self):
        # Close the projection
        self.root.update()
        self.root.destroy()




import numpy as np
from picamera import PiCamera
from time import sleep
from abc import ABC
# if using an environment different to PyCharm
import sys
sys.path.insert(0, '..')
from Camera import Camera


class RaspberryCam(Camera, ABC):
    def __init__(self, exposure=0.1, white_balance=0, auto_focus=True, grayscale=False):
        # Setting and initializing the PiCam
        self.cap = PiCamera()
        # Wait for camera to initialize, otherwise black frames may occur
        sleep(0.5)
        if self.cap is None:
            print("Warning: No RaspberryCam found.")
        # Get framerate and resolution of camera
        fps = self.getFPS()
        resolution = self.getResolution()
        self.grayscale = grayscale
        self.hdr_exposures = None
        super().__init__(exposure, white_balance, auto_focus, fps, resolution, grayscale)

    def getFPS(self):
        # Returns the frame rate
        fps = self.cap.framerate
        return fps

    def setFPS(self, fps):
        # Sets framerate
        self.cap.framerate = fps
        self.fps = fps

    def setAutoGain(self):
        raise NotImplementedError

    def getGain(self):
        raise NotImplementedError

    def setGain(self):
        raise NotImplementedError

    def getResolution(self):
        # Returns a tuple resolution (width, height)
        return self.cap.resolution

    def setResolution(self, resolution):
        # Sets new tuple of resolution
        self.cap.resolution = resolution
        self.resolution = resolution
        print("Camera resolution: ", self.resolution)

    def getImage(self, name='test', saveImage=True, saveNumpy=True, calibration=False):
        # Grab current frame
        # TODO!!!!
        frame = np.empty((self.resolution[1], self.resolution[0], 3), dtype=np.uint8)
        self.cap.capture(frame, 'rgb')
        # TODO: Save to jpg
        # if saveImage:
        #    self.cap.capture('CapturedImages/capture_%i.jpg' % count, 'jpeg')
        if saveNumpy:
            if calibration:
                np.save('CalibrationNumpyData/' + name, frame)
            else:
                np.save('CapturedNumpyData/' + name, frame)

    def setAutoExposure(self):
        # Turn on auto exposure
        self.cap.shutter_speed = 0

    def setExposure(self, exposure):
        # Set exposure value
        self.cap.shutter_speed = exposure
        self.cap.exposure_speed = exposure
        self.exposure = exposure

    def getExposure(self):
        # Returns exposure value
        return self.cap.exposure_speed

    def viewCameraStream(self, time=10):
        # Live preview
        self.cap.start_preview()
        input("Press Enter to continue")
        self.cap.stop_preview()

    def quit_and_close(self):
        # Close camera
        self.cap.close()

    def quit_and_open(self):
        # Close camera
        self.cap.close()
        # Create new capture
        self.cap = PiCamera()

    def getStatus(self):
        if self.cap is None:
            print('Warning: Unable to open Rasperry Cam: ')


from abc import abstractmethod


class CalibrationSession:
    def __init__(self, camera, path='CalibrationImages'):
        # Initialize the camera to calibrate
        self.camera = camera
        # This path points to where images for calibration are stored
        self.path = path

    @abstractmethod
    def capture(self):
        # To override: Image capture procedure for different calibrations
        pass

    @abstractmethod
    def calibrate(self, calibration):
        # To override: Calibration computation
        pass





from abc import abstractmethod


class Projection:
    def __init__(self, resolution, frequency):
        # Tuple Resolution of display/screen in pixels (Width, Height)
        if resolution[0] < resolution[1]:
            self.resolution = reversed(resolution)
        else:
            self.resolution = resolution
        # Frequency Projections
        self.frequency = frequency
        self.calibration = None

    @abstractmethod
    def displayPatterns(self, patterns, camera):
        # To override
        pass

    @abstractmethod
    def getResolution(self):
        # To override
        pass

    @abstractmethod
    def setResolution(self, resolution):
        # To override
        pass

    @abstractmethod
    def quit_and_close(self):
        # To override
        pass

    @abstractmethod
    def getStatus(self):
        # To override
        pass



from abc import ABC
from GhostScan.CaptureSession import CaptureSession
from Projections import Pattern


class DeflectometryCapture(CaptureSession, ABC):
    def __init__(self, camera, projection, image_processing, calibration=None, nph=4):
        self.camera = camera
        self.projection = projection
        # Set sinus pattern as it's frequency depends on the projection resolution
        self.projection.setPattern(Pattern.SinusPattern(projection.resolution))
        self.calibration = calibration
        self.image_processing = image_processing
        # Number of phase shifts
        self.nph = nph

    def capture(self, red=1.0, green=1.0, blue=1.0):
        self.projection.pattern.createSinusXY(self.nph, red=red, blue=blue, green=green)
        # Display patterns, take photos and save as np and jpg
        self.projection.displayPatterns(self.camera)
        self.camera.quit_and_close()

    def compute(self):
        self.image_processing.loadData()
        if self.nph > 4:
            self.image_processing.sinusoidalFitting()
        else:
            self.image_processing.computePhaseMaps()
        self.image_processing.computeNormalMap()
        #self.image_processing.saveTiff()

    def calibrate(self, calibration):
        # Calibrate camera using a calibration object obtained from a Calibration Session
        self.calibration = calibration



from pypylon import pylon
from GhostScan.Camera import Camera
from abc import ABC
import numpy as np
import cv2
import os, time, math
from skimage import io
from io import BytesIO
from IPython.display import clear_output, Image, display, update_display
import PIL
from GhostScan.Cameras.liveDisplay import ptgCamStream, imgShow, patternDisplay
from GhostScan.Cameras.PySpinCapture import PySpinCapture as psc
import matplotlib.pyplot as plt

class Basler(Camera, ABC):

    def __init__(self, exposure=0.01, white_balance=0, auto_focus=False, grayscale=True):
        #  TODO: pylon.FeaturePersistence.Save("test.txt", camera.GetNodeMap())
        # Setting and initializing the Basler camera

        self.cap = pylon.InstantCamera(pylon.TlFactory.GetInstance().CreateFirstDevice())
        self.cap.Open()
        if self.cap is None:
            print('Warning: unable to open external Basler camera')
        # Get framerate and resolution of camera
        fps = self.getFPS()
        resolution = self.getResolution()
        # Init base class
        super().__init__(exposure, white_balance, auto_focus, fps, resolution, grayscale)
        self.hdr_exposures = None

    def setDir(self, directory, sessionName):
        self.directory = directory
        self.sessionName = sessionName
        self.sessionDir = os.path.join(self.directory, self.sessionName)
        print(self.sessionDir)
        if not os.path.exists(self.sessionDir):
            os.makedirs(self.sessionDir)

    def getAutoExposure(self):
        # Returns if auto exposure is enabled
        return self.cap.ExposureAuto.GetValue()

    def setAutoExposure(self):
        # Turn on auto exposure
        self.cap.ExposureAuto.SetValue("Continuous")

    def getFPS(self):
        # Returns the frame rate
        return self.cap.AcquisitionFrameRate.GetValue()

    def setFPS(self, fps):
        # Sets frame rate
        self.cap.AcquisitionFrameRate.SetValue(fps)
        self.fps = fps

    def setAutoGain(self):
        # Set auto gain
        self.cap.GainAuto.SetValue("Once")

    def getGain(self):
        # Returns the set gain value
        return self.cap.Gain.GetValue()

    def setGain(self, gain):
        # Turn off auto gain
        self.cap.GainAuto.SetValue("Off")
        # Set gain value
        self.cap.Gain.SetValue(gain)

    def getResolution(self):
        # Returns a tuple resolution (width, height)
        resolution = (self.cap.Width.GetValue(), self.cap.Height.GetValue())
        return resolution

    def setResolution(self, resolution):
        # Sets the image resolution
        self.cap.Width.SetValue(resolution[0])
        self.cap.Height.SetValue(resolution[1])
        self.resolution = resolution

    def setSingleFrameCapture(self):
        # Set single frame acquisition mode
        self.cap.AcquisitionMode.SetValue('SingleFrame')

    def setHDRExposureValues(self, exposures):
        self.hdr_exposures = exposures

    def setExposure(self, exposure):
        # Set auto exposure off
        self.cap.ExposureAuto.SetValue("Off")
        # Set exposure value in microseconds
        self.cap.ExposureTime.SetValue(exposure)
        self.exposure = exposure

    def getExposure(self):
        # Returns exposure value in microseconds
        return self.cap.ExposureTime.GetValue()

    def getHDRImage(self, name='test', saveImage=True, saveNumpy=True, timeout=5000):
        if self.calibration is None:
            print("Initialize calibration object of camera class first")
        self.cap.StartGrabbingMax(1)
        img = pylon.PylonImage()
        frames = []
        for e in self.hdr_exposures:
            self.setExposure(e)
            while self.cap.IsGrabbing():
                # Grabs photo from camera
                grabResult = self.cap.RetrieveResult(timeout, pylon.TimeoutHandling_ThrowException)
                if grabResult.GrabSucceeded():
                    # Access the image data.
                    frame = grabResult.Array
                    img.AttachGrabResultBuffer(grabResult)
                grabResult.Release()
            frames.append(frame)
        hdr_frame = self.calibration.radio_calib.get_HDR_image(frames, self.hdr_exposures)
        if saveNumpy:
            np.save('CapturedNumpyData/' + name, hdr_frame)
        if saveImage:
            png_frame = (hdr_frame - np.min(hdr_frame)) / (np.max(hdr_frame) - np.min(hdr_frame))
            png_frame *= 255.0
            io.imsave('CapturedImages/' + name + '.PNG', png_frame.astype(np.uint8))
        return hdr_frame

    def getImage(self, name='test', saveImage=True, saveNumpy=True, calibration=False, timeout=5000):
        try:
            # Take and return current camera frame
            self.cap.StartGrabbingMax(1)
            img = pylon.PylonImage()
            while self.cap.IsGrabbing():
                # Grabs photo from camera
                grabResult = self.cap.RetrieveResult(timeout, pylon.TimeoutHandling_ThrowException)
                if grabResult.GrabSucceeded():
                    # Access the image data.
                    frame = grabResult.Array
                    img.AttachGrabResultBuffer(grabResult)
                grabResult.Release()
            # Save if desired
            if saveImage:
                if calibration:
                    filename = 'CalibrationImages/' + name + '.raw'
                    filenamePNG = 'CalibrationImages/' + name + '.PNG'
                    img.Save(pylon.ImageFileFormat_Raw, filename)
                    img.Save(pylon.ImageFileFormat_Png, filenamePNG)
                else:
                    filename = 'CapturedImages/' + name + '.PNG'
                    img.Save(pylon.ImageFileFormat_Png, filename)
            if saveNumpy:
                if calibration:
                    np.save('CalibrationNumpyData/' + name, frame)
                else:
                    np.save('CapturedNumpyData/' + name, frame)
            img.Release()
            self.cap.StopGrabbing()
            return frame
        except SystemError:
            self.quit_and_open()
            return None

    def viewCameraStream(self):
        # Display live view
        while True:
            cv2.namedWindow('Basler Machine Vision Stream', cv2.WINDOW_NORMAL)
            img = self.getImage(saveImage=False, saveNumpy=False)
            print("Max: ", np.max(img))
            print("Min: ", np.min(img))
            cv2.imshow('Basler Machine Vision Stream', img)
            c = cv2.waitKey(1)
            if c != -1:
                # When everything done, release the capture
                cv2.destroyAllWindows()
                break

    def viewCameraStreamSnapshots(self):
        # Display live view
        while True:
            cv2.namedWindow('Basler Machine Vision Stream', cv2.WINDOW_NORMAL)
            img = self.getImage(saveImage=False, saveNumpy=False)
            cv2.imshow('Basler Machine Vision Stream', img)
            c = cv2.waitKey(1)
            if c != -1:
                # When everything done, release the capture
                cv2.destroyAllWindows()
                self.quit_and_open()
                break

    def viewCameraStreamJupyter(self):
        # Live view in a Jupyter Notebook
        try:
            start = self.getImage(saveImage=False, saveNumpy=False)
            g = BytesIO()
            PIL.Image.fromarray(start).save(g, 'jpeg')
            obj = Image(data=g.getvalue())
            dis = display(obj, display_id=True)
            while True:
                img = self.getImage(saveImage=False, saveNumpy=False)
                if img is None:
                    break
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                f = BytesIO()
                PIL.Image.fromarray(img).save(f, 'jpeg')
                obj = Image(data=f.getvalue())
                update_display(obj, display_id=dis.display_id)
                clear_output(wait=True)
        except KeyboardInterrupt:
            self.quit_and_open()
            

    def quit_and_close(self):
        # Close camera
        self.cap.Close()

    def quit_and_open(self):
        # Close camera
        self.cap.Close()
        # Create new capture
        self.cap = pylon.InstantCamera(pylon.TlFactory.GetInstance().CreateFirstDevice())
        self.cap.Open()

    def getStatus(self):
        pylon.FeaturePersistence.Save("Basler_Specs.txt", self.cap.GetNodeMap())


try:

    from Cameras.PySpinCapture import PySpinCapture as psc

    print('1')

except ImportError:

    PySpinCapture = None
    

DISPLAY_HEIGHT = 1080
DISPLAY_WIDTH = 1920
NUM_PATTERN = 7
DEFLECTOMETRY_FREQ = 0.9

class Flir(Camera, ABC):

    def __init__(self, exposure=0.01, white_balance=1, auto_focus=False, grayscale=False):
        self.sessionDir = None

        self._isMonochrome = True

        self._is16bits = True

        self.NumPatterns = NUM_PATTERN
        self.displayWidth = DISPLAY_WIDTH
        self.displayHeight = DISPLAY_HEIGHT

        self.setDefPattern()

        self.Cam = psc(0, self._isMonochrome, self._is16bits)
        self.height = self.Cam.height
        self.width = self.Cam.width

        fps = self.getFPS()

        resolution = self.getResolution()
      
        super().__init__(exposure, white_balance, auto_focus, fps, resolution, grayscale)

        self.hdr_exposures = None
    

    def getImage(self, name='test', saveImage=True, saveNumpy=True, calibration=False, timeout=5000, calibrationName = None):

        try:
            # Take and return current camera frame
            success, img = self.Cam.grabFrame()
            
            # Save if desired
            if saveImage:
                if calibration:
                    #filename = 'CalibrationImages/' + name + '.raw'
                    filenamePNG = 'CalibrationImages/' + name + '.PNG'
                    cv2.imwrite(filenamePNG,cv2.cvtColor(img,cv2.COLOR_BGR2RGB))
                    
                else:
                    filename = 'CapturedImages/' + name + '.PNG'
                    cv2.imwrite(filename, cv2.cvtColor(img, cv2.COLOR_BGR2RGB))

            if saveNumpy:

                if calibration:

                    np.save('CalibrationNumpyData/' + name, img)

                else:

                    np.save('CapturedNumpyData/' + name, img)

            # self.Cam.release()

            return img

        except SystemError:

            self.quit_and_open()

            return None



    def setExposure(self, exposure):

        self.Cam.setExposure(exposure)

        

    def getExposure(self):

        return self.Cam.getExposure()



    def getFPS(self):

        return self.Cam.getFPS()



    def setFPS(self, fps):

        self.Cam.setFPS(fps)



    def setAutoGain(self):

        self.Cam.setCamAutoProperty()



    def getGain(self):

        return self.Cam.getGain()



    def setGain(self, gain):

        self.Cam.setGain(gain)
    
    def captureImage(self, fname):
        if fname:
            path = 'CapturedImages/' + fname + '.png'
            #path = os.path.join(self.sessionDir, fname + ".png")
            flag, img = self.Cam.grabFrame()
            if not flag:
                print("[ERROR]: Didn't get the image!!!")
            else:
                cv2.imwrite(path, img, [cv2.IMWRITE_PNG_COMPRESSION, 0])
        else:
            flag, img = self.Cam.grabFrame()
            print("capture!!!!")
            if not flag:
                print("[ERROR]: Didn't get the image!!!")
                return None
            else:
                return img


    def sinePattern(self, x, y, nu_x, nu_y):
        # Phase Shifts (first entry is white light modulation)
        theta = [0, np.pi / 2, np.pi, 3 / 2 * np.pi]
        [X, Y, Theta] = np.meshgrid(x, y, theta)

        # Calculate Phase Shifts
        phase = (nu_x * X + nu_y * Y) + Theta
        # Simple formula to create fringes between 0 and 1:
        pattern = (np.sin(phase) + 1) / 2

        return pattern
    
    def setPatternScale(self, mode=1):
        patternSize = min(self.displayHeight, self.displayWidth)
        boarderSize = math.floor((max(self.displayHeight, self.displayWidth) - patternSize)/2)

        if mode == 0:
            # Img sequences[7]: {gh, ch, gv, cv, g2h, g2v, black}
            self.NumPatterns = 7
        elif mode == 1:
            # Img sequences[8]: {gh, ch, gv, cv, g2h, g2v, black, full}
            self.NumPatterns = 8
        elif mode == 2:
            # Img sequences[9]: {h1, h2, h3, h4, v1, v2, v3, v4, black}
            self.NumPatterns = 9
        elif mode == 3:
            # Img sequences[9]: {h1, h2, h3, h4, v1, v2, v3, v4, black, full}
            self.NumPatterns = 10
        else:
            self.NumPatterns = 8
            print("[WARNING]: Unrecognizable mode. Using Gradient Pattern!!")
        self.patterns = np.zeros((self.displayHeight, self.displayWidth, self.NumPatterns), dtype=np.float32)

        return patternSize, boarderSize


    def setDefPattern(self):
        #patternSize, boarderSize = self.setPatternScale(2)
        patternSize, boarderSize = self.setPatternScale(2)

        patternSize = max(self.displayHeight, self.displayWidth)
        #patternSize = min(self.displayHeight, self.displayWidth)
        
        # Create spatial coordinates
        x = np.linspace(1, self.displayWidth, self.displayWidth)
        y = np.linspace(1, self.displayHeight, self.displayHeight)

        # Frequencies in x and y direction.
        nu_x =  DEFLECTOMETRY_FREQ * 2 * np.pi / patternSize
        nu_y = DEFLECTOMETRY_FREQ * 2 * np.pi / patternSize

        self.patterns[..., 0:4] = self.sinePattern(x, y, nu_x, 0)
        self.patterns[..., 4:8] = self.sinePattern(x, y, 0, nu_y)
        # self.patterns[..., 8] = 127 * np.ones((DISPLAY_HEIGHT,DISPLAY_WIDTH))

    def captureSeqImages(self):
        window_name = 'projector'
        cv2.namedWindow(window_name, cv2.WND_PROP_FULLSCREEN)

        # If projector is placed right to main screen (see windows properties in your operating system)
        # if the pattern is displayed at the wrong monitor you need to play around with the coordinates here until the image is displayed at the right screen
        cv2.moveWindow(window_name, self.displayWidth, 0)

        cv2.setWindowProperty(window_name, cv2.WND_PROP_FULLSCREEN,cv2.WINDOW_FULLSCREEN)
        cv2.imshow(window_name, self.patterns[..., 0].astype(np.float32))

        if self._isMonochrome:
            if self._is16bits:
                imgs = np.zeros((self.NumPatterns, self.height, self.width), dtype=np.uint16)
            else:
                imgs = np.zeros((self.NumPatterns, self.height, self.width), dtype=np.uint8)
        else:
            if self._is16bits:
                imgs = np.zeros((self.NumPatterns, self.height, self.width, 3), dtype=np.uint16)
            else:
                imgs = np.zeros((self.NumPatterns, self.height, self.width, 3), dtype=np.uint8)

        cv2.waitKey(200)
        # Loop through 
        for i in range(self.NumPatterns):
            if not i==0:
                if i == self.NumPatterns - 1:
                    cur_img = 127 * np.ones((DISPLAY_HEIGHT,DISPLAY_WIDTH))
                    cv2.imshow(window_name, cur_img.astype(np.uint8))
                else:
                    cur_img = self.patterns[... ,i]
                    cv2.imshow(window_name, cur_img.astype(np.float32))

            cv2.waitKey(400)  # ms # delay time
            # wait for 1000 ms ( syncrhonize this with your aquisition time)
            imgs[i, ...] = self.captureImage(fname=None)

            cv2.waitKey(400) # ms # delay time

        # Don't forgot to close all windows at the end of aquisition
        cv2.destroyAllWindows()


        for i in range(self.NumPatterns):
            fname = 'CapturedImages/sequenceImages/' + str(i) + ".png"
            # fname = fpath + '\\' + str(i) + '.png'
            if self._isMonochrome:
                cv2.imwrite(fname, imgs[i, ...], [cv2.IMWRITE_PNG_COMPRESSION, 0])  # RGB to BGR
            else:
                cv2.imwrite(fname, imgs[i, ...][..., ::-1], [cv2.IMWRITE_PNG_COMPRESSION, 0]) #RGB to BGR


    def getResolution(self):

        size = self.Cam.getResolution()

        

        return size



    def setResolution(self, resolution):

        self.Cam.setWidth(resolution[0])

        self.Cam.setHeight(resolution[1])

        

    def getHDRImage(self, name='test', saveImage=True, saveNumpy=True):

        self.Cam.setupHDR()

        imgs = self.Cam.captureHDR()

        if saveNumpy:

            np.save('CapturedNumpyData/' + name, imgs)

        if saveImage:

            png_frame = (imgs - np.min(imgs)) / (np.max(imgs) - np.min(imgs))

            png_frame *= 255.0

            io.imsave('CapturedImages/' + name + '.PNG', png_frame.astype(np.uint8))


    def liveView(self):

        print("[INFO] starting live view thread...")
        viewGrabber = ptgCamStream(self.Cam).start()
        viewShower = imgShow(frame=viewGrabber.read()).start()

        time.sleep(0.1)
        while viewGrabber.running():
            if viewGrabber.stopped or viewShower.stopped:
                viewShower.stop()
                viewGrabber.stop()
                break

            frame = viewGrabber.read()
            viewShower.frame = frame

        viewGrabber.stop()
        viewShower.stop()
        viewGrabber.__exit__()
        viewShower.__exit__()

        viewGrabber.thread.join()
        viewShower.thread.join()
        del viewGrabber
        del viewShower
        cv2.destroyWindow("Live View")

    def viewCameraStream(self,no):
        
        while True:
            _,img = self.Cam.grabFrame()
            cv2.imshow('FLIR camera image ' + str(no), img)
            c = cv2.waitKey(1)
            if c != -1:
                # When everything done, release the capture
                break

            

    def viewCameraStreamJupyter(self):

        # Live view in a Jupyter Notebook
        try:
            start = self.getImage(saveImage=False, saveNumpy=False)
            start = start.astype(np.uint8)
            g = BytesIO()
            obj = Image(data=g.getvalue())
            dis = display(obj, display_id=True)
            while True:
                img = self.getImage(saveImage=False, saveNumpy=False)
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                img = img.astype(np.uint8)
                f = BytesIO()
                PIL.Image.fromarray(img).save(f, 'jpeg')
                obj = Image(data=f.getvalue())
                update_display(obj, display_id=dis.display_id)
        except KeyboardInterrupt:
            self.quit_and_open()

            

    def viewCameraStreamJupyterWindows(self):

        # Live view in a Jupyter Notebook

        try:

            start = self.getImage(saveImage=False, saveNumpy=False)

            g = BytesIO()

            PIL.Image.fromarray(start).save(g, 'jpeg')

            obj = Image(data=g.getvalue())

            dis = display(obj, display_id=True)

            while True:

                img = self.getImage(saveImage=False, saveNumpy=False)

                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

                f = BytesIO()

                PIL.Image.fromarray(img).save(f, 'jpeg')

                obj = Image(data=f.getvalue())

                update_display(obj, display_id=dis.display_id)

        except KeyboardInterrupt:

            self.quit_and_open()

    def displayCalib(self):
        # TODO: Need to use threading so that the GUI won't stuck!
        pattern = cv2.imread("Projections/8_24_checker.png", cv2.IMREAD_GRAYSCALE)
        window_name = 'projector'
        cv2.namedWindow(window_name, cv2.WND_PROP_FULLSCREEN)
        #
        # # If projector is placed right to main screen (see windows properties in your operating system)
        # # if the pattern is displayed at the wrong monitor you need to play around with the coordinates here until the image   is displayed at the right screen
        cv2.moveWindow(window_name, self.displayWidth, 0)
        #
        cv2.setWindowProperty(window_name, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)
        cv2.imshow(window_name, pattern.astype(np.float32))
        #
        # capture checkerboard
        img = self.captureImage(fname=None)
        fname = 'CalibrationImages/Geometric/' + 'test' + ".png"
        cv2.imwrite(fname, img, [cv2.IMWRITE_PNG_COMPRESSION, 0])
        #
        # exit
        k = cv2.waitKey(0)
        if k == 27: # wait for ESC key to exit
            cv2.destroyAllWindows()
         
        

        # patternShower = patternDisplay(displayHeight=self.displayHeight, displayWidth=self.displayWidth, pattern=pattern).start()


    def quit_and_close(self):

        self.Cam.release()



    def quit_and_open(self):

        self.Cam.release()

        self.Cam = psc(0, self._isMonochrome, self._is16bits)



    def getStatus(self):

        raise NotImplementedError



from Camera import Camera
from abc import ABC
import glob
import numpy as np
import cv2
from skimage import io
#from PySpinCapture import PySpinCapture as psc


class NoCamera(Camera, ABC):

    def __init__(self, exposure=0.01, white_balance=0, auto_focus=False, grayscale=True):
        #  TODO: pylon.FeaturePersistence.Save("test.txt", camera.GetNodeMap())
        fps = 0
        resolution = self.getResolution()
        # Init base class
        super().__init__(exposure, white_balance, auto_focus, fps, resolution, grayscale)
        self.hdr_exposures = None

    def getAutoExposure(self):
        raise NotImplementedError

    def setAutoExposure(self):
        raise NotImplementedError

    def getFPS(self):
        raise NotImplementedError

    def setFPS(self, fps):
        raise NotImplementedError

    def setAutoGain(self):
        raise NotImplementedError

    def getGain(self):
        raise NotImplementedError

    def setGain(self, gain):
        raise NotImplementedError

    def getResolution(self):
        return 1200, 1920

    def setResolution(self, resolution):
        self.resolution = resolution

    def setSingleFrameCapture(self):
        raise NotImplementedError

    def setHDRExposureValues(self, exposures):
        raise NotImplementedError

    def setExposure(self, exposure):
        raise NotImplementedError

    def getExposure(self):
        raise NotImplementedError

    def getHDRImage(self, name='test', saveImage=True, saveNumpy=True, timeout=5000):
        raise NotImplementedError

    def getImage(self, name='test', saveImage=True, saveNumpy=True, calibration=False, timeout=5000):
        raise NotImplementedError

    def viewCameraStream(self):
        raise NotImplementedError

    def viewCameraStreamSnapshots(self):
        raise NotImplementedError

    def quit_and_close(self):
        return 0

    def quit_and_open(self):
        return 0

    def getStatus(self):
        raise NotImplementedError


# from Cameras import Webcam, MachineVision, Polarization
from Cameras import MachineVision
from Projections import MainScreen
from CaptureSessions import DeflectometryCapture
from Calibrations import RadiometricCalibration
from CalibrationsSessions import RadiometricCalibSession, RadiometricCalibDisplaySession
#from Reconstructions import DeflectometryReconstruction

#from Visualization import Visualization

nph = 4
# Set-up your camera
# cam = Raspberry.RaspberryCam()
# cam = MachineVision.Basler()
# cam = MachineVision.Flir(exposure=100)
cam = MachineVision.Flir()
cam.setExposure(13000)
cam.setGain(0)
# cam = Webcam.Internal()
# cam.getImage(name = 'test')
# cam.getImage(calibration = True, calibrationName = 'Intrinsic')
# cam.displayCalib()
# intr_calib = IntrinsicCalibration.IntrinsicCalibration()
# intr_calib_session = IntrinsicCalibSession.IntrinsicCalibSession(cam, intr_calib)
# intr_calib_session.capture()
# intr_calib_session.calibrate()
# cam.viewCameraStream()
# cam.setAutoExposure()
# cam.viewCameraStream()
# cam.captureImage(fname="9")
# cam.captureSeqImages()
# cam.liveView()
# Set-up your projector
# projection = MainScreen.Screen()
# Set up image processing
# image_processing = DeflectometryReconstruction.DeflectometryReconstruction()
# Set up your Deflectometry Session
# cap = DeflectometryCapture.DeflectometryCapture(cam, projection, image_processing, 4)
# Capture images
# cap.capture()
# Compute results
# cap.compute()
# Visualize results
'''

vis = Visualization(cap.image_processing)
vis.showPhaseMaps()
vis.showReference()
vis.showAllImages()
vis.showNormals()

'''
'''
intr_calib = IntrinsicCalibration.IntrinsicCalibration()
# Set up the calibration session by passing on the camera and the calibration object
intr_calib_session = IntrinsicCalibSession.IntrinsicCalibSession(cam, intr_calib)
intr_calib_session.capture()
intr_calib_session.calibrate()
intr_calib.load_calibration_data()
'''

# RadiometricCalibration
# Set up the calibration object
radio_calib = RadiometricCalibration.RadiometricCalibration(cam.getResolution())
# Set up the calibration session by passing on the camera and the calibration object
# To set your own exposures please pass them to the RadiometricCalibSession when creating
# radio_calib_session = RadiometricCalibSession.RadiometricCalibSession(cam, radio_calib)
# Capture images at different exposures
# radio_calib_session.capture()


# g, g_n = radio_calib_session.calibrate_HDR(smoothness=500)

# radio_calib.load_calibration_data()
# radio_calib.plotCurve('Grayscale')

radio_calib_display = RadiometricCalibDisplaySession.RadiometricCalibDisplaySession(cam, radio_calib)
# radio_calib_display.captureImages()
# radio_calib_display.avgImagePixel()
# radio_calib_display.calculateDisplayCalibration()
# radio_calib_display.calculateReflectivity()
# after we have got 8 pictures
radio_calib_display.radiometricCalibUndistortImages()


import os, sys, glob
import math
import argparse
import numpy as np
import cv2
from cv2 import aruco
import scipy.spatial
import math
import matplotlib
import matplotlib.pyplot as plt
import GhostScan.Reconstructions.generatephase as gp
from GhostScan.Calibrations.Camera2ScreenCalib import calib
from GhostScan.Reconstructions.normalInWorld import normalInWorld, normalCompare, calAngFeature

#metric_factor = 0.065 / 60 #0.001
# the dimension of glass
#half_length = 0.2875 - 0.005
#half_height = 0.1587 - 0.005
# the dimension of mirror
#half_length = 0.3016 - 0.005
#half_height = 0.1492 - 0.005
# the dimension of round glass samples
#half_length = 0.1524 - 0.005
#half_height = 0.1397 - 0.005

def PhaseMeasuringDeflectometry(rawImgPath, addImgPath, normalPath, metric_factor, cameraMatrix, disCoeffs, ret, rc2s, tc2s, half_length, half_height):
    #load intrinsic matrix and screen2cam transformation
    #sinusoid image path (for calculating phase values)
    #checkerboard image path (for geometric calibration)
    imgFile = addImgPath
    imgUndistortFile = os.path.join(rawImgPath, "0.png")
    #path of output normal map
    normalFile = os.path.join(normalPath, "normal_map.png")
    #cam2screen = np.load("D:/Deflectometry/calibration/Cam2Screen0623_2.npz")
    #rc2s, tc2s = calib(imgFileMirror, intrinsicFile)
    #tc2s = tc2s[0]
    rcamera_dis = rc2s
    tcamera_dis = tc2s
    # additional section: Make sure that the shape of rC2S and tC2S is correct
    tcamera_dis = np.reshape(tcamera_dis, (3,))
    #print("rc2s is ",rcamera_dis)
    #print("tc2s is ",tcamera_dis)   

    #load phasemaps generated by 4-phase shift algorithm
    PhaseMap0, PhaseMap1 = gp.calPhase(imgPath= rawImgPath)
    PhaseMapProjected = np.load('PhaseMapHV.npz')
    PhaseMapHC = PhaseMapProjected.f.arr_0
    PhaseMapVC = PhaseMapProjected.f.arr_1
    #disnplay phase maps
    fig = plt.figure(figsize=(15,10))
    plt.imshow(PhaseMap0)
    plt.title("Horizontal Phase Value")
    plt.colorbar()
    fig.savefig(normalPath + '/Horizontal Phase Value' + '.png')
    plt.show()
    fig = plt.figure(figsize=(15,10))
    plt.plot(PhaseMap0[800, :])
    fig.savefig(normalPath + '/y=800' + '.png')
    plt.show()
    #phaseFile = os.path.join(normalPath, "phase_map_0.png")
    #plt.savefig(phaseFile)
    #cv2.imwrite(phaseFile, PhaseMap0)

    fig = plt.figure(figsize=(15,10))
    plt.imshow(PhaseMap1)
    plt.title("Vertical Phase Value")
    plt.colorbar()
    fig.savefig(normalPath + '/Vertical Phase Value' + '.png')
    plt.show()
    fig = plt.figure(figsize=(15,10))
    plt.plot(PhaseMap1[:, 1000])
    fig.savefig(normalPath + '/x=1000' + '.png')
    plt.show()
    #define board (plane of the object), then estimate the pose of that board using the markers on the object
    allCorners = []
    allIds = []
    decimator = 0
    dictionary = aruco.getPredefinedDictionary(cv2.aruco.DICT_6X6_250)
    img = cv2.imread(imgFile)
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    cornerRefine = cv2.aruco.CORNER_REFINE_CONTOUR
    [markerCorners,markerIDs, rejectedImgPoints] = cv2.aruco.detectMarkers(gray, dictionary, cameraMatrix=cameraMatrix, distCoeff=disCoeffs)


    board_corner = np.array([np.array([[-half_length, -half_height, 0], [-half_length + 0.04, -half_height, 0], [-half_length + 0.04, -half_height + 0.04, 0], [-half_length, -half_height + 0.04, 0]]),
                    np.array([[half_length - 0.04, -half_height, 0], [half_length, -half_height, 0], [half_length, -half_height + 0.04, 0], [half_length - 0.04, -half_height + 0.04, 0]]),
                    np.array([[half_length - 0.04, half_height - 0.04, 0], [half_length, half_height - 0.04, 0], [half_length, half_height, 0], [half_length - 0.04, half_height, 0]]),
                    np.array([[-half_length, half_height - 0.04, 0], [-half_length + 0.04, half_height - 0.04, 0], [-half_length + 0.04, half_height, 0], [-half_length, half_height, 0]]),
                    np.array([[-0.02, -half_height, 0], [0.02, -half_height, 0], [0.02, -half_height + 0.04, 0], [-0.02, -half_height + 0.04, 0]]),
                    np.array([[-half_length, -0.02, 0], [-half_length + 0.04, -0.02, 0], [-half_length + 0.04, 0.02, 0], [-half_length, 0.02, 0]]),
                    np.array([[half_length - 0.04, -0.02, 0], [half_length, -0.02, 0], [half_length, 0.02, 0], [half_length - 0.04, 0.02, 0]])], dtype=np.float32)

    board_id = np.array([[0], [6], [4], [2], [7], [1], [5]], dtype=np.int32)
    board = cv2.aruco.Board_create(board_corner, dictionary, board_id)

    if len(markerCorners) > 0:
        allCorners.append(markerCorners)
        allIds.append(markerIDs)
        cv2.aruco.drawDetectedMarkers(img, markerCorners, markerIDs, [0,200,0])


    rvecs,tvecs,_objpoints = cv2.aruco.estimatePoseSingleMarkers(markerCorners, 0.04, cameraMatrix, disCoeffs, )
    r_mean = np.mean(rvecs,axis=0)
    t_mean = np.mean(tvecs,axis=0)
    r_mirror = None
    t_mirror = None
    ret, r_mirror, t_mirror, = cv2.aruco.estimatePoseBoard(markerCorners, markerIDs, board, cameraMatrix, disCoeffs, r_mirror, t_mirror)
    r_mirror_mat = np.zeros((3,3))
    r_mirror_mat = cv2.Rodrigues(r_mirror,dst= r_mirror_mat,jacobian=None)
    r_mirror_mat = r_mirror_mat[0]
    #print("r mirror mat is ", r_mirror_mat)
    img_axis = aruco.drawAxis(img, cameraMatrix, disCoeffs, r_mirror, t_mirror, 0.04)
    width = 960
    height = int(img.shape[0]*960/img.shape[1])
    smallimg = cv2.resize(img,(width,height))
    smallimg = cv2.cvtColor(smallimg, cv2.COLOR_BGR2RGB)
    plt.figure(figsize=(15,10))
    plt.imshow(smallimg)
    plt.show()
    imsize = img.shape

    #rtcam: camera coordinate system to world coordinate system
    #rtMat: world coordinate system to camera coordinate system
    rtMat = np.hstack((r_mirror_mat, t_mirror))
    rtMat = np.vstack((rtMat, np.array([[0, 0, 0, 1]])))
    rt_cam = np.linalg.inv(rtMat)
    r_cam = rt_cam[0:3, 0:3]
    t_cam = rt_cam[0:3, 3]
    #print("rtMat is",rtMat)
    #print("rtcam in world coor is ", rt_cam)

    #rtdis: display coordinate system to world coordinate system
    #rtdis_inv: world coordinate system to display coordinate system
    t_camdis = np.reshape(tcamera_dis, (tcamera_dis.shape[0],1))
    rtMat_camdis = np.hstack((rcamera_dis, t_camdis))
    rtMat_camdis = np.vstack((rtMat_camdis, np.array([0,0,0,1])))
    rt_dis_inv = np.matmul(rtMat_camdis, rtMat)
    rt_dis = np.linalg.inv(rt_dis_inv)
    r_dis = rt_dis[0:3, 0:3]
    t_dis = rt_dis[0:3,3]
    r_dis_inv = rt_dis_inv[0:3,0:3]
    t_dis_inv = rt_dis_inv[0:3,3]
    #print("rtscreen in world coor is ", rt_dis)
    #print("mirror in screen coor is ",rt_dis_inv)

    #load undistorted image
    img_undistort = cv2.imread(imgUndistortFile)
    gray_undistort = cv2.cvtColor(img_undistort, cv2.COLOR_BGR2GRAY)
    img_undistort_size = gray_undistort.shape

    #reshape t vector
    t_cam = np.reshape(t_cam, (t_cam.shape[0], 1))
    t_dis = np.reshape(t_dis, (t_dis.shape[0], 1))
    t_dis_inv = np.reshape(t_dis_inv, (t_dis_inv.shape[0], 1))

    #cam coor mat:store vectors in camera coordinate system
    #cam world coor mat: store vectors in world coordinate system
    #camera mirror intersect mat: store intersect points between arrival vectors and mirror in world coordinate system
    # display intersect mat: store display intersect points on display coordinate system
    # display intersect mat trans: store display intersect points on world coordinate system
    # reflec mat: store reflectance vectors in world coordinate system 
    # normal mat: store normalized normal vectors in world coordinate system
    # normal mat trans: store normalized normal vectors in camera coordinate system
    # normal mat origin: store mormalized normal vectors in world coordinate system
    img_coor_mat = np.zeros((3, img_undistort_size[0], img_undistort_size[1]))
    cam_coor_mat = np.zeros((3, img_undistort_size[0], img_undistort_size[1]))
    img_coor_mat_rs = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    cam_coor_mat_rs = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    cam_world_coor_mat = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    cam_mirror_intersect_mat = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    reflect_mat = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    display_intersect_mat = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    display_intersect_mat_trans = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    normal_mat = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    normal_mat_trans = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    normal_mat_origin = np.zeros((3, img_undistort_size[0]*img_undistort_size[1]))
    #calculate normal vectors 
    for i in range(img_undistort_size[0]):
        for j in range(img_undistort_size[1]):
            image_coor = np.array([j + 1, i + 1, 1])
            img_coor_mat[:,i,j] = image_coor

    img_coor_mat_rs = np.reshape(img_coor_mat, (3, img_undistort_size[0]*img_undistort_size[1]))
    # calculate arrival vector
    cam_coor_mat_rs = np.matmul(np.linalg.inv(cameraMatrix), img_coor_mat_rs)
    cam_world_coor_mat = np.matmul(r_cam, cam_coor_mat_rs)
    # calculate intersect point
    scale_factor1 = -t_cam[2,0] / cam_world_coor_mat[2,:]
    scale_factor1 = np.reshape(scale_factor1, (1, scale_factor1.shape[0]))
    cam_mirror_intersect_mat = scale_factor1 * cam_world_coor_mat + np.tile(t_cam, (cam_world_coor_mat.shape[1]))
    # reshape phase map (for phase value calculation)
    PhaseMap0_rs = np.reshape(PhaseMap0, (img_undistort_size[0]*img_undistort_size[1],))
    PhaseMap1_rs = np.reshape(PhaseMap1, (img_undistort_size[0]*img_undistort_size[1],))

    # find corresponding position on screen using phase value
    for i in range(img_undistort_size[0]*img_undistort_size[1]):
        phasevalueH = PhaseMap0_rs[i]
        phasevalueV = PhaseMap1_rs[i]
        phasediffH = np.abs(PhaseMapHC - phasevalueH)
        phasediffV = np.abs(PhaseMapVC - phasevalueV)
        h_index = np.argmin(phasediffH)
        v_index = np.argmin(phasediffV)
        h_loc = h_index - img_undistort_size[1]/2
        v_loc = v_index - img_undistort_size[0]/2
        h_loc_metric = h_loc * metric_factor
        v_loc_metric = v_loc * metric_factor
        # x axis adjustment
        h_loc_metric = -h_loc_metric
        display_intersect_point = np.array([h_loc_metric, v_loc_metric, 0])
        display_intersect_mat[:,i] = display_intersect_point

    # calculate arrival and reflectance fector
    display_intersect_mat_trans = np.dot(r_dis, display_intersect_mat) + np.tile(t_dis, (cam_world_coor_mat.shape[1]))
    reflect_mat = display_intersect_mat_trans - cam_mirror_intersect_mat
    arrival_vec_norm = -cam_world_coor_mat
    arrival_vec_norm = arrival_vec_norm / np.linalg.norm(arrival_vec_norm, axis= 0)
    reflect_vec_norm = reflect_mat / np.linalg.norm(reflect_mat, axis= 0)
    # use normalized arrival and reflectance vector to calculate normal fectors, then normalize normal vectors
    normal_mat = arrival_vec_norm + reflect_vec_norm
    normal_mat_origin = normal_mat / np.linalg.norm(normal_mat, axis= 0)
    normal_mat_trans = np.array([normal_mat_origin[1,:], -normal_mat_origin[0,:], -normal_mat_origin[2,:]])

    # restore the shape of matrix and normal vector matrix
    cam_mirror_intersect_mat = np.reshape(cam_mirror_intersect_mat, (3, img_undistort_size[0], img_undistort_size[1]))
    normal_mat_origin = np.reshape(normal_mat_origin, (3, img_undistort_size[0], img_undistort_size[1]))
    normal_mat_trans = np.reshape(normal_mat_trans, (3, img_undistort_size[0], img_undistort_size[1]))
    # the original matrix is (3, 1200, 1900) and we need to transfer it to (1200, 1920, 3) in order to be compatable to normal export codes
    cam_mirror_intersect_rs = np.zeros((img_undistort_size[0], img_undistort_size[1], 3))
    normal_mat_origin_rs = np.zeros((img_undistort_size[0], img_undistort_size[1], 3))
    normal_mat_trans_rs = np.zeros((img_undistort_size[0], img_undistort_size[1], 3))
    for i in range(normal_mat_trans.shape[0]):
        cam_mirror_intersect_rs[:,:,i] = cam_mirror_intersect_mat[i,:,:]
        # Use variance data to filter the invalid normal data (caused by diffused surface)
        if i == 2:
            normal_mat_origin_rs[:,:,i] = normal_mat_origin[i,:,:] 
            normal_mat_trans_rs[:,:,i] = normal_mat_trans[i,:,:] 
        else:
            normal_mat_origin_rs[:,:,i] = normal_mat_origin[i,:,:] 
            # try to magnify x and y elements
            normal_mat_trans_rs[:,:,i] = normal_mat_trans[i,:,:] 

    cam_mirror_intersect_mat = cam_mirror_intersect_rs
    normal_mat_origin = normal_mat_origin_rs
    normal_mat_trans = normal_mat_trans_rs

    #encode the normal vectors into the RGB image
    #np.savez(os.path.join(rootImgPath, "results/intersect_world"),cam_mirror_intersect_mat)
    np.savez(os.path.join(normalPath, "normal_map"), normal_mat_origin)
    #np.savez(os.path.join(rootImgPath, "results/normal_map_trans"), normal_mat_trans)
    normalImage = np.array((normal_mat_trans + 1) / 2.0 * 255, dtype= np.uint8)
    cv2.imwrite(normalFile, cv2.cvtColor(np.array((normal_mat_trans + 1) / 2.0 * 255, dtype= np.uint8), cv2.COLOR_RGB2BGR))
    print("normal map generated.")

    # this part transfer the normal map from camera coordiante into world coordinate
    normalmap_world = normalInWorld(cam_mirror_intersect_mat= cam_mirror_intersect_mat, normal_map= normal_mat_origin, half_length= half_length, half_height = half_height)
    return normalImage, normal_mat_origin, normalmap_world
        

        



from setuptools import setup, find_packages
setup(name='GhostScan',
      version='0.5.25',
      description='three-dimensional imaging and projection ',
      url='https://github.com/NU-ACCESS/GhostScan',
      author='Frank Yang',
      author_email='frankyang2024@u.northwestern.edu',
      license='MIT',
      packages=find_packages(),
      install_requires=['numpy>=1.11',
                        'matplotlib>=1.5',
                        'Pillow >= 7.0.0',
                        'matplotlib >= 3.2.1',
                        'numpy >= 1.18.2',
                        'torch >= 1.4.0',
                        'tifffile >= 2020.6.3',
                        'scipy >= 1.5.2',
                        'wavepy >= 1.0.5',
                        'checkerboard >= 0.2.4',
                        'tqdm >= 4.48.2',
                        'opencv-contrib-python >= 4.4.0',
                        'selenium>=3.141.0',
                        'pandas>=1.1.1',
                        'scikit-image>=0.17.2',
                        'ipython>=7.18.1',
                        'bokeh>=0.12.5'])



from abc import ABC
from GhostScan.CaptureSession import CaptureSession
from GhostScan.Projections import Pattern


class GradientIlluminationCapture(CaptureSession, ABC):
    def __init__(self, camera, projection, image_processing, calibration=None, n=4):
        self.camera = camera
        self.projection = projection
        # Set gradient pattern as it's dimensions depends on the projection resolution
        self.projection.setPattern(Pattern.GradientPattern(projection.resolution))
        self.calibration = calibration
        self.image_processing = image_processing
        # Number of gradient images: default 4 - X&Y x R
        self.n = n

    def capture(self, red=1.0, green=1.0, blue=1.0):
        self.projection.pattern.createGradientXY(self.n, red=red, blue=blue, green=green)
        # Display patterns, take photos and save as np and jpg
        self.projection.displayPatterns(self.camera)

    def compute(self):
        if self.projection.root is not None:
            self.projection.quit_and_close()
        self.image_processing.loadData()
        gamma = self.calibration.radio_calib.gamma
        if self.camera.hdr_exposures is None:
            self.image_processing.computeNormalMapSingle(gamma)
        else:
            self.image_processing.computeNormalMapRadiance(gamma)
        self.image_processing.computeAlbedo()
        #self.image_processing.saveTiff()

    def calibrate(self, calibration):
        # Calibrate camera using a calibration object obtained from a Calibration Session
        self.calibration = calibration
        self.camera.setCalibration(calibration)



from abc import ABC
import numpy as np
from PIL import Image, ImageTk
from GhostScan.Projection import Projection


class GradientPattern(Projection, ABC):
    def __init__(self, resolution, frequency=0):
        super().__init__(resolution, frequency)
        self.patterns = None
        # Set frequency depending on screen resolution

    def createGradientXY(self, n=2, red=1.0, green=1.0, blue=1.0):
        #coeff = [-2.39958185, 7.09471598, -6.94586026, 3.19411896, 0.02044966]
        # Number gradient shifts n:
        # Set up pattern list to store phase shift images (X and Y direction)
        self.patterns = np.zeros((self.resolution[1], self.resolution[0], 3, n*2+1))
        # Create constant pattern
        c = np.ones((self.resolution[1], self.resolution[0]))
        # Create gradient
        x = np.linspace(0, 1, self.resolution[0])
        y = np.linspace(0, 1, self.resolution[1])
        #x = coeff[0] * x ** 4 + coeff[1] * x ** 3 + coeff[2] * x ** 2 + coeff[3] * x + coeff[4]
        #y = coeff[0] * y ** 4 + coeff[1] * y ** 3 + coeff[2] * y ** 2 + coeff[3] * y + coeff[4]
        # Reverse gradient
        xR = np.flipud(x)
        yR = np.flipud(y)
        # Create mesh grid
        [gradientX, gradientY] = np.meshgrid(x, y)
        [gradientXR, gradientYR] = np.meshgrid(xR, yR)

        # Set up pattern
        self.patterns[:, :, 0, 0] = red * gradientX
        self.patterns[:, :, 1, 0] = green * gradientX
        self.patterns[:, :, 2, 0] = blue * gradientX
        self.patterns[:, :, 0, 1] = red * gradientXR
        self.patterns[:, :, 1, 1] = green * gradientXR
        self.patterns[:, :, 2, 1] = blue * gradientXR
        self.patterns[:, :, 0, 2] = red * gradientY
        self.patterns[:, :, 1, 2] = green * gradientY
        self.patterns[:, :, 2, 2] = blue * gradientY
        self.patterns[:, :, 0, 3] = red * gradientYR
        self.patterns[:, :, 1, 3] = green * gradientYR
        self.patterns[:, :, 2, 3] = blue * gradientYR
        self.patterns[:, :, 0, 4] = red * c
        self.patterns[:, :, 1, 4] = green * c
        self.patterns[:, :, 2, 4] = blue * c
        return self.patterns


class SinusPattern(Projection, ABC):
    def __init__(self, resolution, frequency=0):
        super().__init__(resolution, frequency)
        self.x = np.linspace(1, self.resolution[0], self.resolution[0])
        self.y = np.linspace(1, self.resolution[1], self.resolution[1])
        [self.X, self.Y] = np.meshgrid(self.x, self.y)
        self.patterns = None
        # Set frequency depending on screen resolution
        self.frequency = 1 / self.resolution[0]

    def createSinusXY(self, nph, red=1.0, green=1.0, blue=1.0):
        # Number of phase shifts: nph
        # Set up pattern list to store phase shift images (X and Y direction)
        self.patterns = np.zeros((self.resolution[1], self.resolution[0], 3, nph * 2))
        # Set up phase map
        phaseX = self.frequency * self.X
        phaseY = self.frequency * self.Y
        # Loop of number_of_phase_shifts to create sinusoidal patterns in X and Y direction
        for i in range(nph):
            # Calculate Phase Shifts in X and Y direction
            # Simple formula to create fringes between 0 and 1:
            phase_shift = i * 2 * np.pi / nph
            self.patterns[:, :, 0, i] = red * (np.cos(phaseX - phase_shift) + 1) / 2
            self.patterns[:, :, 1, i] = green * (np.cos(phaseX - phase_shift) + 1) / 2
            self.patterns[:, :, 2, i] = blue * (np.cos(phaseX - phase_shift) + 1) / 2
            self.patterns[:, :, 0, i + nph] = red * (np.cos(phaseY - phase_shift) + 1) / 2
            self.patterns[:, :, 1, i + nph] = green * (np.cos(phaseY - phase_shift) + 1) / 2
            self.patterns[:, :, 2, i + nph] = blue * (np.cos(phaseY - phase_shift) + 1) / 2
        return self.patterns


class StepPattern(Projection, ABC):

    def __init__(self, resolution, frequency=0):
        super().__init__(resolution, frequency)
        self.patterns = None
        # Set frequency depending on screen resolution

    def createStep(self, n=50):
        # Number gradient shifts n:
        # Set up pattern list to store phase shift images (X and Y direction)
        self.patterns = np.zeros((self.resolution[1], self.resolution[0], n+1))
        # Create calibrated values
        x = np.linspace(0, 1, 50)
        #coeff1 = [-2.39958185,  7.09471598, -6.94586026,  3.19411896,  0.02044966]
        #y = coeff1[0]*x**4 + coeff1[1]*x**3 + coeff1[2]*x**2 + coeff1[3]*x + coeff1[4]
        # Create constant patterns
        for i in range(n):
            self.patterns[:, :, i] = np.ones((self.resolution[1], self.resolution[0])) * x[i]
        return self.patterns



from abc import ABC
from GhostScan.CalibrationSession import CalibrationSession
import os
import cv2


class IntrinsicCalibSession(CalibrationSession, ABC):
    def __init__(self, camera, intrinsic_calib, path='CalibrationImages', no=20):
        # Set camera
        self.camera = camera
        self.path = path
        self.no = no
        self.intrinsic_calib = intrinsic_calib

    def capture(self):
        # For an intrinsic calibration you are required to image the same Aruco markers from different 20 viewing
        # positions changing in rotation and translation of the camera
        if not os.path.exists('CalibrationImages/Intrinsic/'):
            if not os.path.exists('CalibrationImages'):
                os.mkdir('CalibrationImages')
            os.mkdir('CalibrationImages/Intrinsic/')
        if not os.path.exists('CapturedImages/Intrinsic/'):
            if not os.path.exists('CapturedImages'):
                os.mkdir('CapturedImages')
            os.mkdir('CapturedImages/Intrinsic/')
        if not os.path.exists('CalibrationNumpyData/Intrinsic/'):
            if not os.path.exists('CalibrationNumpyData'):
                os.mkdir('CalibrationNumpyData')
            os.mkdir('CalibrationNumpyData/Intrinsic/')
        if not os.path.exists('CapturedNumpyData/Intrinsic/'):
            if not os.path.exists('CapturedNumpyData'):
                os.mkdir('CapturedNumpyData')
            os.mkdir('CapturedNumpyData/Intrinsic/')
        
        try:
            for i in range(self.no):
                # Display camera stream
                self.camera.viewCameraStream(i)
                # Save images for calibration in .RAW and .PNG format
                filename = 'intr_' + str(i)
                self.camera.getImage(name= 'Intrinsic/' + filename, calibration=True)
                print("Alternate the viewing direction of the camera..." + str(i+1))
            self.camera.quit_and_open()
            cv2.destroyAllWindows()
            cv2.waitKey(1)
            cv2.waitKey(1)
            
        except KeyboardInterrupt:
            self.camera.quit_and_open()
            cv2.destroyAllWindows()
            cv2.waitKey(1)
            cv2.waitKey(1)

    def calibrate(self, imgFolder='CalibrationImages/Intrinsic', imgPattern="*.PNG"):
        # Call calibration in calibration object
        self.intrinsic_calib.calibration(imgFolder, imgPattern)

    def undistort_images(self, imgFolder='CapturedImages', imgPattern="*.PNG"):
        # Call undistort in calibration object
        self.intrinsic_calib.undistort(imgFolder, imgPattern)

    def undistort_npy(self, imgFolder='CapturedNumpyData', imgPattern="*.npy"):
        # Call undistort in calibration object
        self.intrinsic_calib.undistort(imgFolder, imgPattern)





import matplotlib
import matplotlib.pyplot as plt
import numpy as np


class Visualization:
    def __init__(self, image_processing):
        # Initialize image processing object that contains normals, phase maps and others
        self.ip = image_processing

    def showAlbedo(self):
        # Show first frame as reference
        if len(self.ip.albedo.shape) >= 3:
            plt.imshow(self.ip.albedo.astype(np.uint8))
        else:
            plt.imshow(self.ip.albedo, cmap='gray')
        plt.title("Grayscale Reference Frame")
        plt.xlabel("Pixel location x")
        plt.ylabel("Pixel location y")
        plt.show()

    def showPhaseMaps(self):
        # Show phase maps in x and y direction
        n_phase_x = self.ip.diff_x
        n_phase_y = self.ip.diff_y
        # n_phase_x /= n_phase_x.max()
        # n_phase_y /= n_phase_y.max()
        f, ax = plt.subplots(2, 1)
        ax[0].imshow(n_phase_x)
        ax[0].set_title("Phase X")
        ax[0].set_xlabel("Pixel location x")
        ax[0].set_ylabel("Pixel location y")
        ax[1].imshow(n_phase_y)
        ax[1].set_title("Phase Y")
        ax[1].set_xlabel("Pixel location x")
        ax[1].set_ylabel("Pixel location y")
        plt.show()
        
    def showAllImages(self):
        # Show all captured images
        if len(self.ip.albedo.shape) >= 3:
            f, ax = plt.subplots(self.ip.frames_x.shape[3], self.ip.frames_y.shape[3]+1)
            for i in range(self.ip.frames_x.shape[3]):
                ax[i, 0].imshow(self.ip.frames_x[..., i])
                ax[i, 0].set_title("Capture X" + str(i))
                ax[i, 0].set_xlabel("Pixel location x")
                ax[i, 0].set_ylabel("Pixel location y")
                ax[i, 1].imshow(self.ip.frames_y[..., i])
                ax[i, 1].set_title("Capture Y" + str(i))
                ax[i, 1].set_xlabel("Pixel location x")
                ax[i, 1].set_ylabel("Pixel location y")
            ax[0, self.ip.frames_y.shape[3]].imshow(self.ip.frame_reflectivity)
            ax[0, self.ip.frames_y.shape[3]].set_title("Full illumination")
            ax[0, self.ip.frames_y.shape[3]].set_xlabel("Pixel location x")
            ax[0, self.ip.frames_y.shape[3]].set_ylabel("Pixel location y")
            ax[1, self.ip.frames_y.shape[3]].imshow(self.ip.albedo)
            ax[1, self.ip.frames_y.shape[3]].set_title("Mean image")
            ax[1, self.ip.frames_y.shape[3]].set_xlabel("Pixel location x")
            ax[1, self.ip.frames_y.shape[3]].set_ylabel("Pixel location y")
        else:
            f, ax = plt.subplots(self.ip.frames_x.shape[2], self.ip.frames_y.shape[2]+1)
            for i in range(self.ip.frames_x.shape[2]):
                ax[i, 0].imshow(self.ip.frames_x[:, :, i], cmap='gray')
                ax[i, 0].set_title("Phase shift X")
                ax[i, 0].set_xlabel("Pixel location x")
                ax[i, 0].set_ylabel("Pixel location y")
                ax[i, 1].imshow(self.ip.frames_y[:, :, i], cmap='gray')
                ax[i, 1].set_title("Phase shift Y")
                ax[i, 1].set_xlabel("Pixel location x")
                ax[i, 1].set_ylabel("Pixel location y")
            ax[0, self.ip.frames_y.shape[2]].imshow(self.ip.frame_reflectivity, cmap='gray')
            ax[0, self.ip.frames_y.shape[2]].set_title("Full illumination")
            ax[0, self.ip.frames_y.shape[2]].set_xlabel("Pixel location x")
            ax[0, self.ip.frames_y.shape[2]].set_ylabel("Pixel location y")
            ax[1, self.ip.frames_y.shape[2]].imshow(self.ip.albedo, cmap='gray')
            ax[1, self.ip.frames_y.shape[2]].set_title("Mean image")
            ax[1, self.ip.frames_y.shape[2]].set_xlabel("Pixel location x")
            ax[1, self.ip.frames_y.shape[2]].set_ylabel("Pixel location y")
        plt.show()

    def showNormals(self):
        # Show normals as RGB representation
        showN = np.array((self.ip.normals + 1) / 2.0 * 255).astype(int)
        plt.imshow(showN)
        plt.title("Normals")
        plt.xlabel("Pixel location x")
        plt.ylabel("Pixel location y")
        plt.show()

    def showQuiverNormals(self, stride=50):
        # Plot normals as arrows at strided positions
        crop_n_x = self.ip.normals[::stride, ::stride, 0]
        crop_n_y = self.ip.normals[::stride, ::stride, 1]
        # Create a mesh-grid of the image
        x = np.linspace(0, self.ip.normals.shape[1], crop_n_x.shape[1])
        y = np.linspace(0, self.ip.normals.shape[0],  crop_n_x.shape[0])
        X, Y = np.meshgrid(x, y)
        # Plot the normals shown on top of the albedo image
        plt.quiver(X, Y, crop_n_x, crop_n_y, color='g', scale=10)
        plt.imshow(self.ip.albedo, cmap='gray')
        plt.show()



class Calibration:
    def __init__(self, radio_calib=None, intr_calib=None, geo_calib=None):
        # This class simply combines the calibration objects from the different calibration procedures
        self.radio_calib = radio_calib
        self.intr_calib = intr_calib
        self.geo_calib = geo_calib






import os, sys, glob
import math
import argparse
import numpy as np
import cv2
from cv2 import aruco
import scipy.spatial
import math
import matplotlib
import matplotlib.pyplot as plt
import GhostScan.Reconstructions.generatephase as gp
from GhostScan.Calibrations.Camera2ScreenCalib import calib
from GhostScan.Reconstructions.normalInWorld import normalInWorld, normalCompare, calAngFeature


def normal2RGB(normalImage, normal_mat_origin, normalmap_world):
    # Now let's visualize the normal map!
    resultFolder = os.path.join(os.path.join(os.path.join(os.getcwd(), 'CapturedImages'), 'sequenceImages'), 'results')
    fig = plt.figure(figsize=(15,10))
    plt.imshow(normalImage)
    plt.title("Normal Map Encoded in RGB Image")
    fig.savefig(resultFolder + '/Normal Map Encoded in RGB Image' + '.png')
    plt.show()
    # visualize normal map in camera perspective
    norm = matplotlib.colors.Normalize(vmin= -1, vmax = 1)
    norm2 = matplotlib.colors.Normalize(vmin= -1, vmax = 1)
    fig = plt.figure(figsize=(15,10))
    plt.imshow(normal_mat_origin[:,:,0], norm= norm)
    plt.title("X Component")
    plt.colorbar()
    fig.savefig(resultFolder + '/X Component' + '.png')
    plt.show()
    fig = plt.figure(figsize=(15,10))
    plt.imshow(normal_mat_origin[:,:,1], norm= norm)
    plt.title("Y Component")
    plt.colorbar()
    fig.savefig(resultFolder + '/Y Component' + '.png')
    plt.show()
    fig = plt.figure(figsize=(15,10))
    plt.imshow(normal_mat_origin[:,:,2], norm= norm2)
    plt.title("Z Component")
    plt.colorbar()
    fig.savefig(resultFolder + '/Z Component' + '.png')
    plt.show()
    # visualize normal map in world perspective
    fig = plt.figure(figsize=(25,10))
    for i in range(3):
        #plt.figure()
        if i == 2:
            plt.subplot(1,3,i+1)
            plt.imshow(normalmap_world[:,:,i], norm= norm2)
            plt.title("Actual Z")
            plt.tick_params(labelsize=16)
        else:
            plt.subplot(1,3,i+1)
            plt.imshow(normalmap_world[:,:,i], norm= norm)
            if i == 1:
                plt.title("Actual Y")
            else:
                plt.title("Actual X")
            plt.tick_params(labelsize=16)
        cb = plt.colorbar()
        cb.ax.tick_params(labelsize= 16)
    plt.show()
    fig.savefig(resultFolder + '/normal2RGB' + '.png')
        

        



import sys, glob, os
import math
import argparse
import numpy as np
import cv2
from cv2 import aruco
from GhostScan.Calibrations.Camera2ScreenCalib import calib
# from Camera2ScreenCalib import calib
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def geoCalibII(geoCalibImgUndistortFile, displayintersect, displayScaleFactor):
    displayintersectX = displayintersect[0,:,:]
    displayintersectY = displayintersect[1,:,:]
    displayintersectX = -displayintersectX
    displayintersectX = displayintersectX / displayScaleFactor
    displayintersectY = displayintersectY / displayScaleFactor
    rawImg = cv2.imread(geoCalibImgUndistortFile)
    rawImgGray = cv2.cvtColor(rawImg, cv2.COLOR_BGR2GRAY)
    dispImg = cv2.imread("8_24_checker.png")
    dispImgGray = cv2.cvtColor(dispImg, cv2.COLOR_BGR2GRAY)
    dispImgGray = cv2.flip(dispImgGray, 1)
    fig = plt.figure(figsize=(15,10))
    plt.imshow(rawImgGray)
    imgSavePath = os.path.join(os.path.join(os.path.join(os.getcwd(), 'CalibrationImages'), 'Geometric'), 'geoCalibResults')
    fig.savefig(imgSavePath + '/original' + '.png')
    plt.show()
    minX = -dispImgGray.shape[1] / 2
    maxX = dispImgGray.shape[1] / 2
    minY = -dispImgGray.shape[0] / 2
    maxY = dispImgGray.shape[0] / 2
    dispPattern = dispImgGray

    retrieveImage = np.zeros((displayintersectX.shape[0], displayintersectX.shape[1]))
    print(retrieveImage.shape)
    for i in range(retrieveImage.shape[0]):
        for j in range(retrieveImage.shape[1]):
            roundCor = [np.round(displayintersectY[i,j]), np.round(displayintersectX[i,j])]
            if minY < roundCor[0] and roundCor[0] < maxY and minX < roundCor[1] and roundCor[1] < maxX:
                #print(int(roundCor[0] + maxY), int(roundCor[1] + maxX))
                retrieveImage[i,j] = dispPattern[int(roundCor[0] + maxY), int(roundCor[1] + maxX)]
    compareImg = retrieveImage * 0.5 + rawImgGray + 0.5
    fig = plt.figure(figsize=(15,10))
    plt.imshow(retrieveImage)
    fig.savefig(imgSavePath + '/reprojected' + '.png')
    plt.show()
    fig = plt.figure(figsize=(15,10))
    plt.imshow(compareImg)
    fig.savefig(imgSavePath + '/comparation' + '.png')
    plt.show()

    #return display_intersect_mat, cam_mirror_intersect_mat, rcamera_dis, tcamera_dis



import os, sys, glob
import math
import argparse
import numpy as np
import cv2
from cv2 import aruco
import scipy.spatial
import math
import matplotlib.pyplot as plt

IMGPATTERN = "*.png"
PRECISION_IMG=np.dtype(np.float32)
numCam = 8

def readFileList(imgFolder):
    print(os.path.join(imgFolder, IMGPATTERN))
    imgFileList = glob.glob(os.path.join(imgFolder, IMGPATTERN))
    # self.imgFileList = os.listdir(self.imgFolder)
    # self.imgFileList.remove('.DS_Store') # remove system database log
    imgFileList.sort()

    return imgFileList

def processDeflectometry(imgPath):

    imgFileList = readFileList(imgPath)
    resultPath = os.path.join(imgPath, "results")
    if not os.path.exists(resultPath):
        os.makedirs(resultPath)

    assert (len(imgFileList) == 8), ("Images in the directory should be 8. Now is ",len(imgFileList))

    img = cv2.imread(os.path.join(imgPath, imgFileList[0]), cv2.IMREAD_COLOR)
    imgSetDeflectometry = np.zeros((8, img.shape[0], img.shape[1], img.shape[2]), dtype=np.float32)
    imgSetDeflectometry[0, ...] = img

    for i in range(len(imgFileList) - 1):

        img = cv2.imread(os.path.join(imgPath, imgFileList[i + 1]), cv2.IMREAD_COLOR)

        # if i == 0:
        #     imgSetDeflectometry = np.zeros((8, img.shape[0], img.shape[1], img.shape[2]), dtype=np.float32)

        imgSetDeflectometry[i + 1, ...] = img
    return imgSetDeflectometry

def setImgSet(imgSet):

    assert imgSet.shape[0] == numCam, "Deflectometry require 8 different patterns!"

    [height, width, nColor] = imgSet[0, ...].shape
    imgSet = (imgSet / np.max(imgSet)).astype(PRECISION_IMG)
    imgVSet = np.zeros((int(numCam / 2), height, width), dtype=PRECISION_IMG)
    imgHSet = np.zeros((int(numCam / 2), height, width), dtype=PRECISION_IMG)

    for n in range(numCam):
        if n < int(numCam / 2):
            img = cv2.cvtColor(imgSet[n, ...], cv2.COLOR_BGR2GRAY)
            imgVSet[n, ...] = img
        elif n >= int(numCam / 2):
            img = cv2.cvtColor(imgSet[n, ...], cv2.COLOR_BGR2GRAY)
            imgHSet[n - int(numCam / 2), ...] = img
    return imgHSet, imgVSet

def setImgSetVar(imgSet):

    assert imgSet.shape[0] == numCam, "Deflectometry require 8 different patterns!"

    [height, width, nColor] = imgSet[0, ...].shape
    #imgSet = (imgSet / np.max(imgSet)).astype(PRECISION_IMG)
    imgVSet = np.zeros((int(numCam / 2), height, width), dtype=PRECISION_IMG)
    imgHSet = np.zeros((int(numCam / 2), height, width), dtype=PRECISION_IMG)

    for n in range(numCam):
        if n < int(numCam / 2):
            img = cv2.cvtColor(imgSet[n, ...], cv2.COLOR_BGR2GRAY)
            imgVSet[n, ...] = img
        elif n >= int(numCam / 2):
            img = cv2.cvtColor(imgSet[n, ...], cv2.COLOR_BGR2GRAY)
            imgHSet[n - int(numCam / 2), ...] = img
    return imgHSet, imgVSet

def calPhase(imgPath, correctPhaseWrap= 3.8):
    imgset = processDeflectometry(imgPath)
    imgHSet, imgVSet = setImgSet(imgSet= imgset)
    for i in range(2):
        if i == 0:
            imgSet = imgVSet
        else:
            imgSet = imgHSet
        PhaseMap = -1 * np.arctan2((imgSet[0, ...] - imgSet[2, ...]), (imgSet[1, ...] - imgSet[3, ...]))
        if correctPhaseWrap != 0:
            PhaseMap = np.angle(np.exp(1j * (PhaseMap - correctPhaseWrap)))
        if i == 0:
            PhaseMap0 = PhaseMap
        else:
            PhaseMap1 = PhaseMap
    print(imgHSet.shape)

    return PhaseMap0, PhaseMap1   


"""
imgPath = "D:/Deflectometry/pyspin0316/mirror0623"
imgset = processDeflectometry(imgPath= imgPath)
print(imgset.shape)
imghset, imgvset = setImgSet(imgSet= imgset)
print(imghset.shape, imgvset.shape)
p0, p1 = calPhase(imgPath)
plt.imshow(p0)
plt.show()
plt.imshow(p1)
plt.show()
"""





from Cameras import Webcam, MachineVision
from Projections import MainScreen
from CaptureSessions import DeflectometryCapture, GradientShiftingCapture, StepUpCapture
from Calibrations import RadiometricCalibration, IntrinsicCalibration
from CalibrationsSessions import RadiometricCalibSession, IntrinsicCalibSession
from Reconstructions import DeflectometryReconstruction, GradientShiftingReconstruction
from Visualization import Visualization
import Calibration
import numpy as np
import os
import cv2
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

"""
# Set-up your camera
cam = MachineVision.Basler()# Radiometric calib
r_cal = RadiometricCalibration.RadiometricCalibration(cam.getResolution())
cal = RadiometricCalibSession.RadiometricCalibSession(cam, r_cal)
r_cal.load_calibration_data()
hdr_exp = np.array([30000.0, 60000.0, 120000.0, 240000.0])
cam.setHDRExposureValues(hdr_exp)
# Set calibration object
calib = Calibration.Calibration(radio_calib=r_cal)
cam.setCalibration(calib)
"""
cam = MachineVision.Basler()
cam.setExposure(100000.0)
#cam.setAutoGain()
#cam.viewCameraStream()

# Set-up your projector
projection = MainScreen.Screen()

# Set up your Deflectometry Session
cap = StepUpCapture.StepUpCapture(cam, projection)
# Capture images
#cap.capture()
cap.compute()
print("Capturing...")

"""
path='CapturedNumpyData/capture_%i.npy'
frame_0 = np.load(path % 0)
frames = np.zeros((frame_0.shape[0], frame_0.shape[1], 50))
for i in range(0, 50):
    print(i)
    frames[..., i] = np.load(path % i)
print(frames.shape)
g_0 = frames[460:700, 610:700, :]
g_1 = frames[480:720, 730:800, :]
g_2 = frames[500:740, 830:920, :]
g_3 = frames[520:760, 960:1050, :]
for i in range(0, 50):
    g_0[..., i] = frames[460:700, 610:700, i]#/frames[460:700, 610:700, -1]
    g_1[..., i] = frames[480:720, 730:800, i]#/frames[480:720, 730:800, -1]
    g_2[..., i] = frames[500:740, 830:920, i]#/frames[500:740, 830:920, -1]
    g_3[..., i] = frames[520:760, 960:1050, i]#/frames[520:760, 960:1050, -1]

print(g_0.shape)

plt.imshow(g_0[..., 49], cmap='gray')
plt.show()
plt.imshow(g_1[..., 49], cmap='gray')
plt.show()
plt.imshow(g_2[..., 49], cmap='gray')
plt.show()
plt.imshow(g_3[..., 49], cmap='gray')
plt.show()

m_0 = np.mean(g_0, axis=(0, 1))
m_1 = np.mean(g_1, axis=(0, 1))
m_2 = np.mean(g_2, axis=(0, 1))
m_3 = np.mean(g_3, axis=(0, 1))

x = np.linspace(0, 255, 50)
plt.plot(x, m_0, label='White')
plt.plot(x, m_1, label='Flesh')
plt.plot(x, m_2, label='Brown')
plt.plot(x, m_3, label='Black')
plt.xlabel('Projected intensity')
plt.ylabel('Captured mean intensity')
plt.legend()
plt.show()


#cam.setAutoGain()
#cam.viewCameraStream()

"""

"""
# Set-up your projector
#projection = MainScreen.Screen()
resolution = (1920, 1080) #projection.getResolution()
# Create gradient

x = np.linspace(0, 1, resolution[0])
xD = np.linspace(0, resolution[0], resolution[0])
print(x.shape)
print(xD.shape)
y = np.linspace(0, 1, resolution[1])

coeff = [-6.33634985, 10.56707136, -4.06277237,  0.79012877, -0.01104522]
coeff1 = [-2.39958185,  7.09471598, -6.94586026,  3.19411896,  0.02044966]
x1 = coeff[0]*x**4 + coeff[1]*x**3 + coeff[2]*x**2 + coeff[3]*x + coeff[4]
x2 = coeff1[0]*x**4 + coeff1[1]*x**3 + coeff1[2]*x**2 + coeff1[3]*x + coeff1[4]
y1 = coeff[0]*y**3 + coeff[1]*y**2 + coeff[2]*y + coeff[3]
plt.plot(xD, x, label='Linear gradient')
plt.plot(xD, x1, label='Captured gradient')
plt.plot(xD, x2, label='Corrected gradient')
plt.legend()
plt.xlabel('Pixel on display along x-axis')
plt.ylabel('Normalized pixel intensity value')
plt.show()

# Create mesh grid
[gradientX, gradientY] = np.meshgrid(x1, y1)
cv2.imwrite('CalibrationImages/gradientXL.PNG', gradientX*255)
cv2.imwrite('CalibrationImages/gradientYL.PNG', gradientY*255)

cam = MachineVision.Basler()
exp = [1000, 2500, 5000, 7500, 10000, 15000, 20000]
for e in exp:
    cam.setExposure(e)
    cam.viewCameraStream()
    cam.getImage(name=str(e))

"""
"""
# Radiometric Calibration

cam = MachineVision.Basler()
#cam.setExposure(40000)
#exp = 32000
#cam.setExposure(exp)
#cam.viewCameraStream()
# cam = Webcam.Internal()

Exposure = []
ExposureImages = []
files = []
path = '/Users/Merlin/Desktop/CalibrationImages/ExposuresForHDR/capture_4'
for file in os.listdir(path):
    if file.endswith(".npy"):
        files.append(file)
files.sort(key=lambda x: int(x[:-4]))
files.sort()

# We used exposure time as filenames
for filename in files:
    image = np.load(path + '/' + filename)
    filename = os.path.splitext(filename)[0] + '\n'
    Exposure.append(int(filename))
    ExposureImages.append(image)

exps = np.array(Exposure)
imgs = np.array(ExposureImages)
print(exps)
print(imgs.shape)

r_cal = RadiometricCalibration.RadiometricCalibration(cam.getResolution())
cal = RadiometricCalibSession.RadiometricCalibSession(cam, r_cal)
#cal.capture()
exposures = [30, 60, 100, 200, 400, 600, 1000, 1500, 2000, 4000, 6000, 8000, 10000, 14000, 19000,
                              24000, 31000, 39000, 49000, 60000]
#g, g_n = cal.calibrate_HDR(smoothness=1000)
#r_cal.get_camera_response(1000, exposures)
r_cal.load_calibration_data()
#r_cal.CRC(1000)
#cal.load_calibration()
r_cal.get_HDR_image()
#z, zs = r_cal.CRC(1000)
#r_cal.get_camera_response()
#r_cal.get_HDR_image()
"""
"""
#cam.getImage()
img, g_exp = cal.calibrate_image(exp)
# use the created array to output your multiple images. In this case I have stacked 4 images vertically
f, axarr = plt.subplots(2,2)
i0 = axarr[0,0].imshow(img[0])
axarr[0,0].set_title('Normalized radiance map 0')
divider = make_axes_locatable(axarr[0,0])
cax = divider.append_axes("right", size="5%", pad=0.05)
plt.colorbar(i0, cax=cax)
i1 = axarr[0,1].imshow(img[1])
axarr[0,1].set_title('Normalized radiance map 1')
divider = make_axes_locatable(axarr[0,1])
cax = divider.append_axes("right", size="5%", pad=0.05)
plt.colorbar(i1, cax=cax)
i2 = axarr[1,0].imshow(img[2])
axarr[1,0].set_title('Normalized radiance map 2')
divider = make_axes_locatable(axarr[1,0])
cax = divider.append_axes("right", size="5%", pad=0.05)
plt.colorbar(i2, cax=cax)
i3 = axarr[1,1].imshow(img[3])
axarr[1,1].set_title('Normalized radiance map 3')
divider = make_axes_locatable(axarr[1,1])
cax = divider.append_axes("right", size="5%", pad=0.05)
plt.colorbar(i3, cax=cax)
plt.show()

"""
"""
np.save('CalibrationImages/Distorted/capture_0.npy', np.squeeze(img[0]))
np.save('CalibrationImages/Distorted/capture_1.npy', np.squeeze(img[1]))
np.save('CalibrationImages/Distorted/capture_2.npy', np.squeeze(img[2]))
np.save('CalibrationImages/Distorted/capture_3.npy', np.squeeze(img[3]))


cam.quit_and_close()

org = cv2.imread('/Users/Merlin/Desktop/Gradients/Screen/7500.PNG')
plt.figure()

#subplot(r,c) provide the no. of rows and columns
f, axarr = plt.subplots(2, 1)
# use the created array to output your multiple images. In this case I have stacked 4 images vertically
axarr[0].imshow(org, cmap='gray')
axarr[1].imshow(img[0], cmap='gray')
plt.show()
"""

# Intrinsic calibration

cam = MachineVision.Basler()
# cam.setAutoExposure()
intr_calib = IntrinsicCalibration.IntrinsicCalibration()
calibo = IntrinsicCalibSession.IntrinsicCalibSession(cam, intr_calib)
# calibo.capture()
calibo.calibrate()

